import logging
from pathlib import Path
import sys
import warnings
from app_model.types import (
    Action,
    ToggleRule,
    KeyCode,
    KeyMod,
    KeyChord,
    StandardKeyBinding,
)
from himena._descriptors import SaveToPath, NoNeedToSave
from himena.consts import MenuId, StandardType
from himena.utils.html import html_to_plain_text
from himena.widgets import MainWindow, SubWindow
from himena.types import (
    ClipboardDataModel,
    WindowState,
    WidgetDataModel,
)
from himena._app_model._context import AppContext as _ctx
from himena._app_model.actions._registry import ACTIONS, SUBMENUS
from himena import _utils, _providers
from himena.exceptions import Cancelled
from himena.workflow import LocalReaderMethod

_LOGGER = logging.getLogger(__name__)

EDIT_GROUP = "00_edit"
PROPERTY_GROUP = "01_property"
STATE_GROUP = "21_state"
MOVE_GROUP = "22_move"
ZOOM_GROUP = "40_zoom"
EXIT_GROUP = "99_exit"
_CtrlK = KeyMod.CtrlCmd | KeyCode.KeyK
_CtrlShift = KeyMod.CtrlCmd | KeyMod.Shift


@ACTIONS.append_from_fn(
    id="show-whats-this",
    title="What Is This Widget?",
    menus=[{"id": MenuId.WINDOW, "group": EXIT_GROUP}],
    enablement=_ctx.num_sub_windows > 0,
)
def show_whats_this(ui: MainWindow) -> None:
    """Show the docstring of the current widget."""
    if window := ui.current_window:
        if doc := getattr(window.widget, "__doc__", ""):
            doc_formatted = _utils.doc_to_whats_this(doc)
            ui._backend_main_window._add_whats_this(doc_formatted, style="markdown")


@ACTIONS.append_from_fn(
    id="show-workflow-graph",
    title="Show Workflow Graph",
    menus=[
        {"id": MenuId.WINDOW, "group": EXIT_GROUP},
        {"id": MenuId.TOOLBAR, "group": EXIT_GROUP},
    ],
    enablement=_ctx.num_sub_windows > 0,
    need_function_callback=True,
    icon="hugeicons:workflow-circle-05",
)
def show_workflow_graph(model: WidgetDataModel) -> WidgetDataModel:
    """Show the workflow graph of the current window."""
    workflow = model.workflow
    return WidgetDataModel(
        value=workflow.deep_copy(),
        type=StandardType.WORKFLOW,
        title=f"Workflow of {model.title}",
        save_behavior_override=NoNeedToSave(),
        extension_default=".workflow.json",
        workflow=workflow,
    )


@ACTIONS.append_from_fn(
    id="open-last-closed-window",
    title="Open Last Closed Window",
    menus=[{"id": MenuId.WINDOW, "group": EXIT_GROUP}],
    keybindings=[{"primary": _CtrlShift | KeyCode.KeyT}],
)
def open_last_closed_window(ui: MainWindow) -> WidgetDataModel:
    """Open the last closed window."""
    if last := ui._history_closed.pop_last():
        path, plugin = last
        store = _providers.ReaderStore().instance()
        model = store.run(path=path, plugin=plugin)
        model.workflow = LocalReaderMethod(
            output_model_type=model.type, plugin=plugin, path=path
        ).construct_workflow()
        return model
    warnings.warn("No window to reopen", UserWarning, stacklevel=2)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="close-window",
    title="Close Window",
    icon="material-symbols:tab-close-outline",
    menus=[
        {"id": MenuId.WINDOW, "group": EXIT_GROUP},
    ],
    keybindings=[StandardKeyBinding.Close],
    enablement=_ctx.num_sub_windows > 0,
)
def close_current_window(ui: MainWindow) -> None:
    """Close the selected sub-window."""
    i_tab = ui.tabs.current_index
    if i_tab is None:
        raise Cancelled
    tab = ui.tabs[i_tab]
    i_window = tab.current_index
    if i_window is None:
        raise Cancelled
    _LOGGER.info(f"Closing window {i_window} in tab {i_tab}")
    tab[i_window]._close_me(ui, ui._instructions.confirm)


@ACTIONS.append_from_fn(
    id="duplicate-window",
    title="Duplicate Window",
    enablement=_ctx.is_active_window_supports_to_model,
    menus=[{"id": MenuId.WINDOW, "group": EDIT_GROUP}],
    keybindings=[{"primary": KeyChord(_CtrlK, _CtrlShift | KeyCode.KeyD)}],
    need_function_callback=True,
)
def duplicate_window(win: SubWindow) -> WidgetDataModel:
    """Duplicate the selected sub-window."""
    # NOTE: whether this copies the internal data depends on the widget.
    model = win.to_model()
    update = {
        "save_behavior_override": NoNeedToSave(),
        "force_open_with": _utils.get_widget_class_id(type(win.widget)),
    }
    if model.title is not None:
        model = model.with_title_numbering()
    return model.model_copy(update=update)


@ACTIONS.append_from_fn(
    id="rename-window",
    title="Rename Window",
    menus=[
        {"id": MenuId.WINDOW, "group": EDIT_GROUP},
    ],
    enablement=_ctx.num_sub_windows > 0,
    keybindings=[{"primary": KeyChord(_CtrlK, KeyCode.F2)}],
)
def rename_window(ui: MainWindow) -> None:
    """Rename the title of the window."""
    i_tab = ui.tabs.current_index
    if i_tab is None:
        return None
    if (i_win := ui._backend_main_window._current_sub_window_index(i_tab)) is not None:
        ui._backend_main_window._rename_window_at(i_tab, i_win)
    return None


@ACTIONS.append_from_fn(
    id="copy-path-to-clipboard",
    title="Copy Path To Clipboard",
    menus=[{"id": MenuId.WINDOW, "group": EDIT_GROUP}],
    enablement=_ctx.num_sub_windows > 0,
    keybindings=[{"primary": KeyChord(_CtrlK, _CtrlShift | KeyCode.KeyC)}],
)
def copy_path_to_clipboard(ui: MainWindow) -> ClipboardDataModel:
    """Copy the path of the current window to the clipboard."""
    if window := ui.current_window:
        if isinstance(sv := window.save_behavior, SaveToPath):
            return ClipboardDataModel(text=str(sv.path))
        else:
            warnings.warn(
                "Window does not have the source path.", UserWarning, stacklevel=2
            )
    else:
        warnings.warn("No window is focused.", UserWarning, stacklevel=2)
    raise Cancelled


@ACTIONS.append_from_fn(
    id="copy-data-to-clipboard",
    title="Copy Data To Clipboard",
    menus=[
        {"id": MenuId.WINDOW, "group": EDIT_GROUP},
    ],
    enablement=(_ctx.num_sub_windows > 0) & _ctx.is_active_window_supports_to_model,
    keybindings=[{"primary": KeyChord(_CtrlK, KeyMod.CtrlCmd | KeyCode.KeyC)}],
)
def copy_data_to_clipboard(model: WidgetDataModel) -> ClipboardDataModel:
    """Copy the data of the current window to the clipboard."""

    if model.is_subtype_of(StandardType.TEXT):
        if model.is_subtype_of(StandardType.HTML):
            return ClipboardDataModel(
                text=html_to_plain_text(model.value), html=model.value
            )
        else:
            return ClipboardDataModel(text=model.value)
    elif model.is_subtype_of(StandardType.IMAGE):
        return ClipboardDataModel(image=model.value)
    raise ValueError(f"Cannot convert {model.type} to a clipboard data.")


@ACTIONS.append_from_fn(
    id="minimize-window",
    title="Minimize Window",
    menus=[{"id": MenuId.WINDOW_RESIZE, "group": STATE_GROUP}],
    keybindings=[{"primary": KeyChord(_CtrlK, KeyMod.CtrlCmd | KeyCode.DownArrow)}],
    enablement=_ctx.num_sub_windows > 0,
)
def minimize_current_window(win: SubWindow) -> None:
    """Minimize the window"""
    win.state = WindowState.MIN


@ACTIONS.append_from_fn(
    id="maximize-window",
    title="Maximize Window",
    menus=[{"id": MenuId.WINDOW_RESIZE, "group": STATE_GROUP}],
    enablement=_ctx.num_sub_windows > 0,
    keybindings=[{"primary": KeyChord(_CtrlK, KeyMod.CtrlCmd | KeyCode.UpArrow)}],
)
def maximize_current_window(win: SubWindow) -> None:
    win.state = WindowState.MAX


@ACTIONS.append_from_fn(
    id="toggle-full-screen",
    title="Toggle Full Screen",
    menus=[{"id": MenuId.WINDOW_RESIZE, "group": STATE_GROUP}],
    keybindings=[{"primary": KeyCode.F11}],
    enablement=_ctx.num_sub_windows > 0,
)
def toggle_full_screen(win: SubWindow) -> None:
    if win.state is WindowState.FULL:
        win.state = WindowState.NORMAL
    else:
        win.state = WindowState.FULL


@ACTIONS.append_from_fn(
    id="unset-anchor",
    title="Unanchor Window",
    menus=[MenuId.WINDOW_ANCHOR],
    enablement=_ctx.num_sub_windows > 0,
)
def unset_anchor(win: SubWindow) -> None:
    """Unset the anchor of the window if exists."""
    win.anchor = None


@ACTIONS.append_from_fn(
    id="anchor-window-top-left",
    title="Anchor Window To Top-Left Corner",
    menus=[MenuId.WINDOW_ANCHOR],
    enablement=_ctx.num_sub_windows > 0,
)
def anchor_at_top_left(win: SubWindow) -> None:
    """Anchor the window at the top-left corner of the current window position."""
    win.anchor = "top-left"


@ACTIONS.append_from_fn(
    id="anchor-window-top-right",
    title="Anchor Window To Top-Right Corner",
    menus=[MenuId.WINDOW_ANCHOR],
    enablement=_ctx.num_sub_windows > 0,
)
def anchor_at_top_right(win: SubWindow) -> None:
    """Anchor the window at the top-right corner of the current window position."""
    win.anchor = "top-right"


@ACTIONS.append_from_fn(
    id="anchor-window-bottom-left",
    title="Anchor Window To Bottom-Left Corner",
    menus=[MenuId.WINDOW_ANCHOR],
    enablement=_ctx.num_sub_windows > 0,
)
def anchor_at_bottom_left(win: SubWindow) -> None:
    """Anchor the window at the bottom-left corner of the current window position."""
    win.anchor = "bottom-left"


@ACTIONS.append_from_fn(
    id="anchor-window-bottom-right",
    title="Anchor Window To Bottom-Right Corner",
    menus=[MenuId.WINDOW_ANCHOR],
    enablement=_ctx.num_sub_windows > 0,
)
def anchor_at_bottom_right(win: SubWindow) -> None:
    """Anchor the window at the bottom-right corner of the current window position."""
    win.anchor = "bottom-right"


@ACTIONS.append_from_fn(
    id="window-expand",
    title="Expand (+20%)",
    enablement=_ctx.num_sub_windows > 0,
    menus=[{"id": MenuId.WINDOW_RESIZE, "group": ZOOM_GROUP}],
    keybindings=[StandardKeyBinding.ZoomIn],
)
def window_expand(win: SubWindow) -> None:
    """Expand (increase the size of) the current window."""
    win._set_rect(win.rect.resize_relative(1.2, 1.2))


@ACTIONS.append_from_fn(
    id="window-shrink",
    title="Shrink (-20%)",
    enablement=_ctx.num_sub_windows > 0,
    menus=[{"id": MenuId.WINDOW_RESIZE, "group": ZOOM_GROUP}],
    keybindings=[StandardKeyBinding.ZoomOut],
)
def window_shrink(win: SubWindow) -> None:
    """Shrink (reduce the size of) the current window."""
    win._set_rect(win.rect.resize_relative(1 / 1.2, 1 / 1.2))


@ACTIONS.append_from_fn(
    id="full-screen-in-new-tab",
    title="Full Screen In New Tab",
    enablement=_ctx.num_sub_windows > 0,
    menus=[{"id": MenuId.WINDOW, "group": EDIT_GROUP}],
)
def full_screen_in_new_tab(ui: MainWindow) -> None:
    """Move the selected sub-window to a new tab and make it full screen."""
    if win := ui.current_window:
        ui.add_tab(win.title)
        index_new = len(ui.tabs) - 1
        ui.move_window(win, index_new)
        win.state = WindowState.FULL
        ui.tabs.current_index = index_new


@ACTIONS.append_from_fn(
    id="reveal-in-explorer",
    title="Reveal In Explorer",
    enablement=_ctx.num_sub_windows > 0,
    menus=[{"id": MenuId.WINDOW, "group": EDIT_GROUP}],
)
def reveal_in_explorer(win: SubWindow):
    from subprocess import Popen

    if isinstance(win.save_behavior, SaveToPath) and win.save_behavior.path.exists():
        path = win.save_behavior.path
    elif isinstance(source := win.to_model().source, Path) and source.exists():
        path = source
    else:
        raise ValueError("Could not determine the source file of the window.")

    if sys.platform == "darwin":
        if path.is_dir():
            Popen(["open", "-R", str(path)])
        else:
            Popen(["open", "-R", str(path.parent)])
    elif sys.platform == "win32":
        Popen(["explorer", "/select,", str(path)])
    elif sys.platform == "linux":
        Popen(["xdg-open", str(path.parent)])
    else:
        raise NotImplementedError(f"Platform {sys.platform} is not supported")


_CtrlAlt = KeyMod.CtrlCmd | KeyMod.Alt


@ACTIONS.append_from_fn(
    id="align-window-left",
    title="Align Window To Left",
    enablement=_ctx.num_sub_windows > 0,
    menus=[MenuId.WINDOW_ALIGN],
    keybindings=[{"primary": _CtrlAlt | KeyCode.LeftArrow}],
)
def align_window_left(ui: MainWindow) -> None:
    """Align the window to the left edge of the tab area."""
    if window := ui.current_window:
        window._set_rect(window.rect.align_left(ui.area_size))


@ACTIONS.append_from_fn(
    id="align-window-right",
    title="Align Window To Right",
    enablement=_ctx.num_sub_windows > 0,
    menus=[MenuId.WINDOW_ALIGN],
    keybindings=[{"primary": _CtrlAlt | KeyCode.RightArrow}],
)
def align_window_right(ui: MainWindow) -> None:
    """Align the window to the right edge of the tab area."""
    if window := ui.current_window:
        window._set_rect(window.rect.align_right(ui.area_size))


@ACTIONS.append_from_fn(
    id="align-window-top",
    title="Align Window To Top",
    enablement=_ctx.num_sub_windows > 0,
    menus=[MenuId.WINDOW_ALIGN],
    keybindings=[{"primary": _CtrlAlt | KeyCode.UpArrow}],
)
def align_window_top(ui: MainWindow) -> None:
    """Align the window to the top edge of the tab area."""
    if window := ui.current_window:
        window._set_rect(window.rect.align_top(ui.area_size))


@ACTIONS.append_from_fn(
    id="align-window-bottom",
    title="Align Window To Bottom",
    enablement=_ctx.num_sub_windows > 0,
    menus=[MenuId.WINDOW_ALIGN],
    keybindings=[{"primary": _CtrlAlt | KeyCode.DownArrow}],
)
def align_window_bottom(ui: MainWindow) -> None:
    """Align the window to the bottom edge of the tab area."""
    if window := ui.current_window:
        window._set_rect(window.rect.align_bottom(ui.area_size))


@ACTIONS.append_from_fn(
    id="align-window-center",
    title="Align Window To Center",
    enablement=_ctx.num_sub_windows > 0,
    menus=[MenuId.WINDOW_ALIGN],
    keybindings=[{"primary": _CtrlAlt | KeyCode.Space}],
)
def align_window_center(ui: MainWindow) -> None:
    """Align the window to the center of the tab area."""
    if window := ui.current_window:
        window._set_rect(window.rect.align_center(ui.area_size))


def toggle_editable(win: SubWindow) -> None:
    win.is_editable = not win.is_editable


def toggle_track_modification(win: SubWindow) -> None:
    win._set_modification_tracking(not win._data_modifications.track_enabled)


ACTIONS.append(
    Action(
        id="window-toggle-editable",
        title="Window Editable",
        callback=toggle_editable,
        enablement=_ctx.is_subwindow_focused,
        menus=[{"id": MenuId.WINDOW, "group": PROPERTY_GROUP}],
        toggled=ToggleRule(condition=_ctx.is_active_window_editable),
    )
)

ACTIONS.append(
    Action(
        id="window-toggle-track-modifications",
        title="Track User Modifications",
        callback=toggle_editable,
        enablement=_ctx.is_subwindow_focused,
        menus=[{"id": MenuId.WINDOW, "group": PROPERTY_GROUP}],
        toggled=ToggleRule(condition=_ctx.is_active_window_track_modification),
    )
)


SUBMENUS.append_from(
    id=MenuId.WINDOW,
    submenu=MenuId.WINDOW_RESIZE,
    title="Resize",
    enablement=_ctx.num_sub_windows > 0,
    group=MOVE_GROUP,
)
SUBMENUS.append_from(
    id=MenuId.WINDOW,
    submenu=MenuId.WINDOW_ALIGN,
    title="Align",
    enablement=_ctx.num_sub_windows > 0,
    group=MOVE_GROUP,
)
SUBMENUS.append_from(
    id=MenuId.WINDOW,
    submenu=MenuId.WINDOW_ANCHOR,
    title="Anchor",
    enablement=_ctx.num_sub_windows > 0,
    group=MOVE_GROUP,
)



from __future__ import annotations

from dataclasses import dataclass, field, is_dataclass, fields
from functools import partial
import random
import json
import logging
from copy import deepcopy
from pathlib import Path
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Mapping,
    NamedTuple,
    Sequence,
    TypeVar,
    overload,
    TYPE_CHECKING,
)

from pydantic_compat import BaseModel

from app_model import Action
from app_model.types import SubmenuItem, KeyBindingRule
from app_model.expressions import BoolOp

from himena._app_model import AppContext as ctx
from himena.consts import MenuId
from himena import _utils
from himena.types import WidgetDataModel
from himena.utils.collections import OrderedSet
from himena.plugins import _utils as _plugins_utils

if TYPE_CHECKING:
    from himena._app_model import HimenaApplication

    KeyBindingsType = str | KeyBindingRule | Sequence[str] | Sequence[KeyBindingRule]
    PluginConfigType = Any

_F = TypeVar("_F", bound=Callable)
_T = TypeVar("_T")
_LOGGER = logging.getLogger(__name__)


class PluginConfigTuple(NamedTuple):
    title: str
    config: PluginConfigType
    config_class: type

    def as_dict(self) -> dict[str, Any]:
        """Convert the config into a normalized dictionary."""
        config = self.config
        if isinstance(config, Mapping):
            out = deepcopy(config)
        elif is_dataclass(config):
            out = {}
            for _f in fields(config):
                out[_f.name] = {"value": getattr(config, _f.name), **_f.metadata}
                # NOTE: "annotation" is not always serializable so it is not allowed
                # here.
        elif isinstance(config, BaseModel):
            out = {}
            for _fname, _finfo in config.model_fields.items():
                options = {}
                if tooltip := _finfo.description:
                    options["tooltip"] = tooltip
                options.update(_finfo.metadata)
                out[_f] = {"value": getattr(config, _fname), **options}
        else:
            raise TypeError(
                "Plugin config type must be dict, dataclass or pydantic.BaseModel, but "
                f"got {config!r}"
            )
        return out


@dataclass
class ReproduceArgs:
    """Command and its arguments to reproduce a user modification."""

    command_id: str
    with_params: dict[str, Any] = field(
        default_factory=dict
    )  # values must be serializable


@dataclass
class AppTip:
    short: str
    long: str


class AppActionRegistry:
    _global_instance: AppActionRegistry | None = None

    def __init__(self):
        self._actions: dict[str, Action] = {}
        self._actions_dynamic: set[str] = set()
        self._submenu_titles: dict[str, str] = {
            MenuId.FILE_NEW: "New ...",
            MenuId.TOOLS_DOCK: "Dock widgets",
        }
        self._submenu_groups: dict[str, str] = {
            MenuId.FILE_NEW: "00_new",
            MenuId.TOOLS_DOCK: "00_dock",
        }
        self._submenu_order: dict[str, int] = {}
        self._installed_plugins: list[str] = []
        self._plugin_default_configs: dict[str, PluginConfigTuple] = {}
        self._modification_trackers: dict[str, Callable[[_T, _T], ReproduceArgs]] = {}
        self._app_tips: list[AppTip] = []
        self._try_load_app_tips()

    @classmethod
    def instance(cls) -> AppActionRegistry:
        """Get the global instance of the registry."""
        if cls._global_instance is None:
            cls._global_instance = cls()
        return cls._global_instance

    def _try_load_app_tips(self) -> None:
        tip_path = Path(__file__).parent.parent / "resources" / "tips.json"
        try:
            with tip_path.open("r") as f:
                out = json.load(f)
            for each in out:
                self._app_tips.append(
                    AppTip(
                        short=each.get("short", ""),
                        long=each.get("long", ""),
                    )
                )
        except Exception as e:
            _LOGGER.error("Failed to load app tips: %s", e)

    def pick_a_tip(self) -> AppTip | None:
        """Pick a random tip."""
        if len(self._app_tips) == 0:
            return None
        tip = random.choice(self._app_tips)
        return AppTip(short=tip.short, long=tip.long)

    def add_action(self, action: Action, is_dynamic: bool = False) -> None:
        """Add an action to the registry."""
        id_ = action.id
        if id_ in self._actions:
            raise ValueError(f"Action ID {id_} already exists.")
        self._actions[id_] = action
        if is_dynamic:
            self._actions_dynamic.add(id_)

    @property
    def installed_plugins(self) -> list[str]:
        """List of modules or python paths that are installed as plugins."""
        return self._installed_plugins

    def iter_actions(self, app: HimenaApplication) -> Iterator[Action]:
        for id_, action in self._actions.items():
            if id_ not in app.commands:
                yield action

    def submenu_title(self, id: str) -> str:
        """Get the title of a submenu."""
        if title := self._submenu_titles.get(id):
            return title
        return id.split("/")[-1].title()

    def submenu_group(self, id: str) -> str | None:
        """Get the group of a submenu."""
        return self._submenu_groups.get(id, None)

    def submenu_order(self, id: str) -> str | None:
        return self._submenu_order.get(id, None)

    @property
    def submenu_titles(self) -> dict[str, str]:
        return self._submenu_titles

    def install_to(
        self,
        app: HimenaApplication,
        actions: list[Action] | None = None,
    ) -> list[str]:
        """Install actions to the application.

        This method automatically adds submenus if they are not already exists, and
        returns the list of added root menu IDs. Note that this does NOT updates the
        GUI menubar and toolbar.
        """
        # look for existing menu items
        if actions is None:
            actions = list(self.iter_actions(app))
        existing_menu_ids = {_id.value for _id in MenuId if "/" not in _id.value}
        for menu_id, menu in app.menus:
            existing_menu_ids.add(menu_id)
            for each in menu:
                if isinstance(each, SubmenuItem):
                    existing_menu_ids.add(each.submenu)

        added_menu_ids = OrderedSet[str]()
        for action in actions:
            if action.menus is not None:
                ids = [a.id for a in action.menus]
                added_menu_ids.update(ids)

        # add submenus if not exists
        to_add: list[tuple[str, SubmenuItem]] = []
        new_menu_ids: list[str] = []

        for place in added_menu_ids - existing_menu_ids:
            place_components = place.split("/")
            if len(place_components) == 1:
                new_menu_ids.append(place)
            for i in range(1, len(place_components)):
                menu_id = "/".join(place_components[:i])
                submenu = "/".join(place_components[: i + 1])
                if submenu in existing_menu_ids:
                    continue
                title = self.submenu_title(submenu)
                group = self.submenu_group(submenu)
                order = self.submenu_order(submenu)
                item = SubmenuItem(
                    title=title, submenu=submenu, group=group, order=order
                )
                to_add.append((menu_id, item))

        app.register_actions(actions)
        app.menus.append_menu_items(to_add)
        app._dynamic_command_ids.update(self._actions_dynamic)
        return new_menu_ids


def norm_menus(menus: str | Sequence[str]) -> list[str]:
    if isinstance(menus, str):
        return [menus]
    return list(menus)


def _norm_menus_with_group(menus: str | Sequence[str], group: str) -> list[dict]:
    return [{"id": menu, "group": group} for menu in norm_menus(menus)]


def configure_submenu(
    submenu_id: str | Iterable[str],
    title: str | None = None,
    *,
    group: str | None = None,
    order: int | None = None,
) -> None:
    """Register a configuration for submenu(s).

    Parameters
    ----------
    submenu_id : str or iterable of str
        Submenu ID(s) to configure.
    title : str, optional
        Specify the title of the submenu.
    group : str, optional
        Specify the group ID of the submenu.
    """
    if isinstance(submenu_id, str):
        submenu_id = [submenu_id]
    for sid in submenu_id:
        if title is not None:
            AppActionRegistry.instance()._submenu_titles[sid] = title
        if group is not None:
            AppActionRegistry.instance()._submenu_groups[sid] = group
        if order is not None:
            AppActionRegistry.instance()._submenu_order[sid] = order


@overload
def register_function(
    *,
    menus: str | Sequence[str] = "plugins",
    title: str | None = None,
    types: str | Sequence[str] | None = None,
    enablement: BoolOp | None = None,
    keybindings: Sequence[KeyBindingRule] | None = None,
    run_async: bool = False,
    command_id: str | None = None,
    group: str | None = None,
    icon: str | None = None,
    palette: bool = True,
) -> None: ...


@overload
def register_function(
    func: _F,
    *,
    menus: str | Sequence[str] = "plugins",
    title: str | None = None,
    types: str | Sequence[str] | None = None,
    enablement: BoolOp | None = None,
    keybindings: Sequence[KeyBindingRule] | None = None,
    run_async: bool = False,
    command_id: str | None = None,
    group: str | None = None,
    icon: str | None = None,
    palette: bool = True,
) -> _F: ...


def register_function(
    func=None,
    *,
    menus="plugins",
    title=None,
    types=None,
    enablement=None,
    keybindings=None,
    run_async=False,
    command_id=None,
    group=None,
    icon=None,
    palette=True,
):
    """Register a function as a callback of a plugin action.

    This function can be used either as a decorator or a simple function.

    Parameters
    ----------
    func : callable, optional
        Function to register as an action.
    menus : str or sequence of str, default "plugins"
        Menu(s) to add the action. Submenus are separated by `/`.
    title : str, optional
        Title of the action. Name of the function will be used if not given.
    types: str or sequence of str, optional
        The `type` parameter(s) allowed as the WidgetDataModel. If this parameter
        is given, action will be grayed out if the active window does not satisfy
        the listed types.
    enablement: Expr, optional
        Expression that describes when the action will be enabled. As this argument
        is a generalized version of `types` argument, you cannot use both of them.
    run_async : bool, default False
        If true, the function will be executed asynchronously. Note that if the function
        updates the GUI, running it asynchronously may cause issues.
    command_id : str, optional
        Command ID. If not given, the function qualname will be used.
    group : str, optional
        Group name to which this command belongs. This parameter follows the app-model
        rule.
    icon : str, optional
        Iconify icon key to use for the action.
    palette : bool, default True
        If true, the action will be added to the command palette.
    """

    def _inner(f: _F) -> _F:
        action = make_action_for_function(
            f,
            menus=menus,
            title=title,
            types=types,
            enablement=enablement,
            keybindings=keybindings,
            run_async=run_async,
            command_id=command_id,
            group=group,
            icon=icon,
            palette=palette,
        )
        AppActionRegistry.instance().add_action(action)
        return f

    return _inner if func is None else _inner(func)


register_hidden_function = partial(register_function, menus=[], palette=False)
"""A shorthand for `register_function(menus=[], palette=False)`."""


def make_action_for_function(
    f: Callable,
    *,
    menus="plugins",
    title=None,
    types=None,
    enablement=None,
    keybindings=None,
    run_async: bool = False,
    command_id: str | None = None,
    group: str | None = None,
    icon: str | None = None,
    palette: bool = True,
):
    types, enablement, menus = _norm_register_function_args(types, enablement, menus)
    kbs = normalize_keybindings(keybindings)
    if title is None:
        _title = f.__name__.replace("_", " ").title()
    else:
        _title = title
    _enablement = enablement
    if _utils.has_widget_data_model_argument(f):
        _enablement = _plugins_utils.expr_and(
            _enablement, ctx.is_active_window_supports_to_model
        )

    if inner_widget_class := _utils.get_subwindow_type_arg(f):
        # function is annotated with SubWindow[W]. Use W for enablement.
        widget_id = _utils.get_widget_class_id(inner_widget_class)
        _enablement = _plugins_utils.expr_and(
            _enablement, ctx.active_window_widget_id == widget_id
        )

    _id = command_id_from_func(f, command_id)
    if group is not None:
        menus_normed = _norm_menus_with_group(menus, group)
    elif isinstance(command_id, str) and ":" in command_id:
        group = command_id.rsplit(":", maxsplit=1)[0]
        menus_normed = _norm_menus_with_group(menus, group)
    else:
        menus_normed = menus
    return Action(
        id=_id,
        title=_title,
        tooltip=tooltip_from_func(f),
        callback=_utils.make_function_callback(
            f, command_id=_id, title=_title, run_async=run_async
        ),
        menus=menus_normed,
        enablement=_enablement,
        keybindings=kbs,
        icon=icon,
        icon_visible_in_menu=False,
        palette=palette,
    )


def _norm_register_function_args(
    types: str | Sequence[str] | None,
    enablement: BoolOp | None,
    menus: str | Sequence[str] | None,
) -> tuple[list[str], BoolOp | None, list[str]]:
    if isinstance(types, str):
        _types = [types]
    elif types is None:
        _types = []
    else:
        _types = types
    if len(_types) > 0:
        type_enablement = _plugins_utils.types_to_expression(_types)
        if enablement is None:
            enablement = type_enablement
        else:
            enablement = enablement & type_enablement

    # Registered functions that are specific to a certain model type will also be
    # added to the model menu tool button in the top-right corner of the sub window.
    # To efficiently make QModelMenu, the "menus" attribute of each type of function
    # will be reallocated from "/model_menu/XYZ" to "/model_menu:TYPE/XYZ".
    menu_out: list[str] = []
    model_menu_found = False
    pref = "/model_menu"
    reg = AppActionRegistry.instance()
    for menu in norm_menus(menus):
        # if "/model_menu/..." submenu is specified by the user, reallocate it.
        if _plugins_utils.is_model_menu_prefix(menu):
            _, _, other = menu.split("/", maxsplit=2)
            _LOGGER.debug("Reallocated: %r ", menu)
            for _type in _types:
                new_place = f"{pref}:{_type}/{other}"
                menu_out.append(new_place)
                _LOGGER.debug("  ---> %r", new_place)
            if menu in reg._submenu_titles:
                _LOGGER.debug("Submenu reallocated: %r ", menu)
                rest = menu[len(pref) :]
                for _type in _types:
                    new_place = f"{pref}:{_type}{rest}"
                    reg._submenu_titles[new_place] = reg._submenu_titles[menu]
                    _LOGGER.debug("  ---> %r", new_place)
            model_menu_found = True
        else:
            menu_out.append(menu)
    if not model_menu_found:
        # if "/model_menu/..." submenu is not specified by the user, they will be added
        # under the root of the model menu.
        for _type in _types:
            new_place = f"{pref}:{_type}"
            menu_out.append(new_place)
            _LOGGER.debug("Menu added: %r", new_place)
    return _types, enablement, menu_out


def normalize_keybindings(keybindings):
    if isinstance(keybindings, str):
        kbs = [KeyBindingRule(primary=keybindings)]
    elif isinstance(keybindings, Sequence):
        kbs = []
        for kb in keybindings:
            if isinstance(kb, str):
                kbs.append(KeyBindingRule(primary=kb))
            elif isinstance(kb, Mapping):
                kbs.append(KeyBindingRule(**kb))
            elif isinstance(kb, KeyBindingRule):
                kbs.append(kb)
            else:
                raise TypeError(f"{kb!r} not allowed as a keybinding.")
    elif keybindings is None:
        kbs = keybindings
    else:
        raise TypeError(f"{keybindings!r} not allowed as keybindings.")
    return kbs


def command_id_from_func(func: Callable, command_id: str) -> str:
    """Make a command ID from a function.

    If function `my_func` is defined in a module `my_module`, the command ID will be
    `my_module:my_func`. If `command_id` is given, it will be used instead of `my_func`.
    """
    if command_id is None:
        _id = getattr(func, "__qualname__", str(func))
    else:
        _id = command_id
    return _id


def tooltip_from_func(func: Callable) -> str | None:
    if doc := getattr(func, "__doc__", None):
        tooltip = str(doc)
    else:
        tooltip = None
    return tooltip


@overload
def register_conversion_rule(
    func: _F,
    type_from: str,
    type_to: str,
    *,
    keybindings: KeyBindingsType | None = None,
    command_id: str | None = None,
) -> _F: ...
@overload
def register_conversion_rule(
    type_from: str,
    type_to: str,
    *,
    keybindings: KeyBindingsType | None = None,
    command_id: str | None = None,
) -> Callable[[_F], _F]: ...


def register_conversion_rule(*args, **kwargs):
    """Register a function as a conversion rule.

    A conversion rule will be added to the model menu under "Convert > ..." submenu.
    Essentially, this method does nothing more than `register_function`, but using this
    registration method is recommended for the sake of clarity and following the
    conventions.

    ```python
    from himena.plugins import register_conversion_rule

    @register_conversion_rule("text", "table")
    def convert_text_to_table(model: WidgetDataModel) -> WidgetDataModel:
        ...  # convert the data type
    ```
    """
    if len(args) == 0:
        no_func = True
    else:
        if isinstance(args[0], str):
            return register_conversion_rule(None, *args, **kwargs)
        no_func = args[0] is None

    def inner(func):
        annot = getattr(func, "__annotations__", {})
        annot.setdefault("return", WidgetDataModel)
        func.__annotations__ = annot
        action = _make_conversion_rule(func, *args, **kwargs)
        AppActionRegistry.instance().add_action(action)
        return func

    return inner if no_func else inner(args[0])


def _make_conversion_rule(
    func: Callable,
    type_from: str,
    type_to: str,
    *,
    keybindings: KeyBindingsType | None = None,
    command_id: str | None = None,
):
    kbs = normalize_keybindings(keybindings)
    _id = command_id_from_func(func, command_id)
    title = f"Convert {type_from} to {type_to}"
    type_from_main, *type_from_subs = type_from.split(".")
    enablement = ctx.active_window_model_type == type_from_main
    if len(type_from_subs) > 0:
        enablement &= ctx.active_window_model_subtype_1 == type_from_subs[0]
    if len(type_from_subs) > 1:
        enablement &= ctx.active_window_model_subtype_2 == type_from_subs[1]
    if len(type_from_subs) > 2:
        enablement &= ctx.active_window_model_subtype_3 == type_from_subs[2]
    return Action(
        id=_id,
        title=title,
        tooltip=tooltip_from_func(func),
        callback=_utils.make_function_callback(func, command_id=_id, title=title),
        menus=[{"id": f"/model_menu:{type_from}/convert", "group": "conversion"}],
        enablement=enablement,
        keybindings=kbs,
    )


def register_modification_tracker(
    type: str,
) -> Callable[[Callable[[_T, _T], ReproduceArgs]], Callable[[_T, _T], ReproduceArgs]]:
    """Register a modification tracker."""
    reg = AppActionRegistry.instance()
    if type in reg._modification_trackers:
        raise ValueError(f"Modification tracker for {type} already exists.")

    def inner(fn):
        reg._modification_trackers[type] = fn
        return fn

    return inner


def add_default_status_tip(
    short: str,
    long: str,
) -> None:
    """Add a status tip that will be randomly shown in the status bar."""
    reg = AppActionRegistry.instance()
    reg._app_tips.append(AppTip(short=str(short), long=str(long)))



from __future__ import annotations
from pathlib import Path
from typing import Any, overload

from himena.types import ClipboardDataModel
from himena.widgets import current_instance
from contextlib import suppress


def set_status_tip(text: str, duration: float = 10.0) -> None:
    """Set a status tip to the current main window for duration (second)."""

    with suppress(Exception):
        ins = current_instance()
        ins.set_status_tip(text, duration=duration)
    return None


def get_clipboard() -> ClipboardDataModel:
    """Get the current clipboard data."""
    return current_instance().clipboard


@overload
def set_clipboard(
    *,
    text: str | None = None,
    html: str | None = None,
    image: Any | None = None,
    files: list[str | Path] | None = None,
    interanal_data: Any | None = None,
) -> None: ...


@overload
def set_clipboard(model: ClipboardDataModel, /) -> None: ...


def set_clipboard(model=None, **kwargs) -> None:
    """Set data to clipboard."""
    ins = current_instance()
    if model is not None:
        if kwargs:
            raise TypeError("Cannot specify both model and keyword arguments")
        ins.clipboard = model
    else:
        ins.set_clipboard(**kwargs)
    return None


def notify(text: str, duration: float = 5.0) -> None:
    """Show a notification popup in the bottom right corner."""
    ins = current_instance()
    ins._backend_main_window._show_notification(text, duration)
    return None


def append_result(item: dict[str, Any], /) -> None:
    """Append a new result to the result stack."""
    ins = current_instance()
    ins._backend_main_window._append_result(item)
    return None


# def subprocess_run(command_args, /, *args, blocking: bool = True, **kwargs):
#     """Run a subprocess command."""
#     import subprocess

#     if isinstance(command_args, str):
#         command_args_normed = command_args
#     else:
#         # first check all the types
#         for arg in command_args:
#             if not isinstance(arg, (str, WidgetDataModel)):
#                 raise TypeError(f"Invalid argument type: {type(arg)}")
#         command_args_normed = []
#         for arg in command_args:
#             if isinstance(arg, str):
#                 command_args_normed.append(arg)
#             elif isinstance(arg, WidgetDataModel):
#                 arg.write_to_directory(...)
#                 command_args_normed.append(...)
#             else:
#                 raise RuntimeError("Unreachable code")
#     if blocking:
#         return subprocess.run(command_args_normed, *args, **kwargs)
#     else:
#         return subprocess.Popen(command_args_normed, *args, **kwargs)



from __future__ import annotations
from functools import partial
import inspect
from types import FunctionType
from typing import Any, Callable
import ast
from inspect import getsource

from qtpy import QtWidgets as QtW, QtCore, QtGui

from himena.consts import StandardType, MonospaceFontFamily
from himena.plugins import validate_protocol
from himena.standards.model_meta import FunctionMeta
from himena.types import WidgetDataModel
from himena.style import Theme
from ._text_base import QMainTextEdit


class QFunctionEdit(QtW.QWidget):
    """Widget for a Python function.

    A function can be compiled from a text edit widget.
    """

    __himena_widget_id__ = "builtins:QFunctionEdit"
    __himena_display_name__ = "Built-in Function Editor"

    def __init__(self):
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(2, 2, 2, 2)
        _label = "<code>f(x) = </code>"
        layout.addWidget(QtW.QLabel(_label))

        self._main_text_edit = QMainTextEdit()
        self._main_text_edit.setReadOnly(True)
        layout.addWidget(self._main_text_edit)

        self._parameter_edit = QtW.QWidget()
        self._parameter_layout = QtW.QFormLayout(self._parameter_edit)
        self._parameter_layout.setContentsMargins(0, 0, 0, 0)
        self._parameter_widgets: list[QPythonLiteralLineEdit] = []
        layout.addWidget(self._parameter_edit)

        self._model_type = StandardType.FUNCTION
        self._func_orig: Callable | None = None
        self._has_source_code = False
        self._control = QFunctionEditControl()

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        if not callable(func := model.value):
            raise TypeError(f"Input value must be callable, got {type(func)}.")
        if isinstance(func, partial):
            self._pfunc = func
            _func = func.func
            _func_orig = func
            args = func.args
            keywords = func.keywords
            sig = inspect.signature(func)
            bound = sig.bind_partial(*args, **keywords)
            for _ in range(len(bound.arguments) - self._parameter_layout.count()):
                edit = QPythonLiteralLineEdit()
                self._parameter_layout.addRow(edit)
                self._parameter_widgets.append(edit)
            for _ in range(self._parameter_layout.count() - len(bound.arguments)):
                self._parameter_layout.takeAt(0)
                self._parameter_widgets.pop(0)
            for ith, (key, value) in enumerate(bound.arguments.items()):
                edit = self._parameter_widgets[ith]
                edit.setLabel(key)
                edit.setValue(value)
        else:
            _func = _func_orig = func
        # try to get the source
        code_text: str | None = None
        if isinstance(meta := model.metadata, FunctionMeta):
            code_text = meta.source_code
        if code_text is None and isinstance(_func, FunctionType):
            try:
                code_text = getsource(_func)
            except Exception:
                # local function etc.
                code_text = None
        self._func_orig = _func_orig
        if code_text:
            self._main_text_edit.setPlainText(code_text)
            self._main_text_edit.syntax_highlight("python")
            self._has_source_code = True
        else:
            self._main_text_edit.setPlainText(repr(_func))
            self._main_text_edit.syntax_highlight(None)
            self._has_source_code = False
        self._control._type_label.setText(_function_type_repr(_func_orig))
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        if self._has_source_code:
            code = self._main_text_edit.toPlainText()
        else:
            code = None
        return WidgetDataModel(
            value=self._func_orig,
            type=self.model_type(),
            metadata=FunctionMeta(source_code=code),
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def control_widget(self) -> QtW.QWidget:
        return self._control

    @validate_protocol
    def is_editable(self) -> bool:
        return self._parameter_edit.isEnabled()

    @validate_protocol
    def set_editable(self, editable: bool):
        return self._parameter_edit.setEnabled(editable)

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 280, 200 + 28 * self._parameter_layout.count()

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        text_edit = self._main_text_edit
        if theme.is_light_background():
            text_edit._code_theme = "default"
        else:
            text_edit._code_theme = "native"
        text_edit.syntax_highlight(text_edit._language)

    def setFocus(self):
        self._main_text_edit.setFocus()


class QFunctionEditControl(QtW.QWidget):
    def __init__(self):
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._type_label = QtW.QLabel("")
        self._type_label.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        layout.addWidget(QtW.QWidget(), 10)
        layout.addWidget(self._type_label)


class QPythonLiteralLineEdit(QtW.QWidget):
    """Line edit in param_name = XYZ format"""

    def __init__(self):
        super().__init__()
        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(3, 0, 3, 0)
        self._label = QtW.QLabel("")
        self._label.setFixedWidth(60)
        self._label.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        layout.addWidget(self._label)
        self._value_edit = QtW.QLineEdit()
        layout.addWidget(self._value_edit)
        self.setFont(QtGui.QFont(MonospaceFontFamily))

    def value(self) -> Any | None:
        text = self._value_edit.text().strip()
        if text == "":
            return None
        return ast.literal_eval(text)

    def setValue(self, value: Any | None):
        if value is None:
            self._value_edit.setText("")
        else:
            self._value_edit.setText(repr(value))

    def label(self) -> str:
        return self._label.text()

    def setLabel(self, label: str):
        self._label.setText(label + " = ")


def _function_type_repr(f) -> str:
    if isinstance(f, partial):
        return f"functools.partial of {_function_type_repr(f.func)}"
    ftype = type(f)
    return f"{ftype.__module__}.{ftype.__name__}"



from __future__ import annotations

from typing import Callable, overload, TypeVar, TYPE_CHECKING
from cmap import Color
from himena.standards import plotting as hplt

if TYPE_CHECKING:
    from matplotlib import pyplot as plt
    from mpl_toolkits import mplot3d as plt3d

_CONVERSION_RULES: dict[
    type[hplt.BasePlotModel], Callable[[hplt.BasePlotModel, plt.Axes], None]
] = {}

_F = TypeVar("_F", bound=Callable)


@overload
def register_plot_model(
    model_class: type[hplt.BasePlotModel],
    rule: _F,
) -> _F: ...
@overload
def register_plot_model(
    model_class: type[hplt.BasePlotModel],
    rule: None,
) -> Callable[[_F], _F]: ...


def register_plot_model(
    model_class: type[hplt.BasePlotModel],
    rule: Callable[[hplt.BasePlotModel, plt.Axes], None] | None = None,
):
    """Register a matplotlib-specific conversion rule for a plot model."""

    def inner(f):
        _CONVERSION_RULES[model_class] = f
        return f

    return inner if rule is None else inner(rule)


def convert_plot_model(model: hplt.BasePlotModel, ax: plt.Axes):
    if model.__class__ in _CONVERSION_RULES:
        return _CONVERSION_RULES[model.__class__](model, ax)
    raise ValueError(f"Unsupported plot model: {model!r}")


@register_plot_model(hplt.Scatter)
def _add_scatter(model: hplt.Scatter, ax: plt.Axes):
    kwargs = {}
    if model.size is not None:
        kwargs["s"] = model.size**2
    if model.face.color is not None:
        kwargs["c"] = Color(model.face.color).hex
    if model.edge.color is not None:
        kwargs["edgecolors"] = model.edge.color
    ax.scatter(
        model.x, model.y, marker=model.symbol, linewidths=model.edge.width,
        hatch=model.face.hatch, linestyle=model.edge.style or "-", label=model.name,
        **kwargs,
    )  # fmt: skip


@register_plot_model(hplt.Line)
def _add_line(model: hplt.Line, ax: plt.Axes):
    ax.plot(
        model.x, model.y, color=model.edge.color, linewidth=model.edge.width,
        linestyle=model.edge.style, label=model.name,
    )  # fmt: skip


@register_plot_model(hplt.Bar)
def _add_bar(model: hplt.Bar, ax: plt.Axes):
    kwargs = {}
    if model.bar_width is not None:
        kwargs["width"] = model.bar_width
    ax.bar(
        model.x, model.y, color=model.face.color, hatch=model.face.hatch,
        bottom=model.bottom, edgecolor=model.edge.color, **kwargs,
        label=model.name, linewidth=model.edge.width, linestyle=model.edge.style,
    )  # fmt: skip


@register_plot_model(hplt.Histogram)
def _add_hist(model: hplt.Histogram, ax: plt.Axes):
    _add_band(model.to_band(), ax)


@register_plot_model(hplt.ErrorBar)
def _add_errorbar(model: hplt.ErrorBar, ax: plt.Axes):
    ax.errorbar(
        model.x, model.y, xerr=model.x_error, yerr=model.y_error,
        capsize=model.capsize, color=model.edge.color, linewidth=model.edge.width,
        linestyle=model.edge.style, label=model.name,
    )  # fmt: skip


@register_plot_model(hplt.Band)
def _add_band(model: hplt.Band, ax: plt.Axes):
    if model.orient == "vertical":
        func = ax.fill_between
    else:
        func = ax.fill_betweenx
    func(
        model.x, model.y0, model.y1, color=model.face.color, hatch=model.face.hatch,
        edgecolor=model.edge.color, linewidth=model.edge.width,
        linestyle=model.edge.style or "-", label=model.name,
    )  # fmt: skip


@register_plot_model(hplt.Span)
def _add_span(model: hplt.Span, ax: plt.Axes):
    if model.orient == "vertical":
        func = ax.axvspan
    else:
        func = ax.axhspan
    func(
        model.start,
        model.end,
        color=model.face.color,
        hatch=model.face.hatch,
        edgecolor=model.edge.color,
        linewidth=model.edge.width,
        linestyle=model.edge.style,
        label=model.name,
    )


@register_plot_model(hplt.Texts)
def _add_texts(model: hplt.Texts, ax: plt.Axes):
    ha, va = _ANCHOR_MAP[model.anchor]
    for x, y, text in zip(model.x, model.y, model.texts):
        ax.text(
            x, y, text, color=model.color, fontsize=model.size, clip_on=True,
            label=model.name, rotation=model.rotation, ha=ha, va=va,
        )  # fmt: skip


@register_plot_model(hplt.Scatter3D)
def _add_scatter_3d(model: hplt.Scatter3D, ax: plt3d.Axes3D):
    if model.size is not None:
        s = model.size**2
    else:
        s = 20
    ax.scatter(
        model.x, model.y, model.z, s=s, c=Color(model.face.color).hex,
        marker=model.symbol, linewidths=model.edge.width, hatch=model.face.hatch,
        edgecolors=model.edge.color, linestyle=model.edge.style or "-",
        label=model.name,
    )  # fmt: skip


@register_plot_model(hplt.Line3D)
def _add_line_3d(model: hplt.Line3D, ax: plt3d.Axes3D):
    ax.plot(
        model.x, model.y, model.z, color=model.edge.color, linewidth=model.edge.width,
        linestyle=model.edge.style, label=model.name,
    )  # fmt: skip


@register_plot_model(hplt.Surface3D)
def _add_surface_3d(model: hplt.Surface3D, ax: plt3d.Axes3D):
    ax.plot_surface(
        model.x, model.y, model.z, color=model.face.color, edgecolor=model.edge.color,
        linewidth=model.edge.width or 0.0, linestyle=model.edge.style or "-",
        label=model.name,
    )  # fmt: skip


@register_plot_model(hplt.Mesh3D)
def _add_mesh_3d(model: hplt.Mesh3D, ax: plt3d.Axes3D):
    ax.plot_trisurf(
        model.vertices[:, 0], model.vertices[:, 1], model.vertices[:, 2],
        triangles=model.face_indices, color=model.face.color, edgecolor=model.edge.color,
        linewidth=model.edge.width, linestyle=model.edge.style, label=model.name,
    )  # fmt: skip


_ANCHOR_MAP = {
    "center": ("center", "center"),
    "left": ("left", "center"),
    "right": ("right", "center"),
    "top": ("center", "top"),
    "bottom": ("center", "bottom"),
    "top_left": ("left", "top"),
    "top_right": ("right", "top"),
    "bottom_left": ("left", "bottom"),
    "bottom_right": ("right", "bottom"),
}



from himena_builtins.qt.output._widget import get_widget
from pytestqt.qtbot import QtBot
import logging


def test_stdout(qtbot: QtBot):
    widget = get_widget()._widget
    qtbot.addWidget(widget)
    assert widget._stdout.toPlainText() == ""
    print("Hello")
    assert widget._stdout.toPlainText() == "Hello\n"


def test_logger(qtbot: QtBot):
    widget = get_widget()._widget
    qtbot.addWidget(widget)
    assert widget._logger.toPlainText() == ""
    logger = logging.getLogger("test")
    logger.warning("Hello")
    assert widget._logger.toPlainText() == "WARNING: Hello\n"



"""Builtin standard output plugin."""

from typing import TYPE_CHECKING
from dataclasses import dataclass
from himena.plugins import register_dock_widget_action, config_field
from himena.consts import MenuId

if TYPE_CHECKING:
    from himena.widgets import MainWindow


@dataclass
class OutputConfig:
    """Configuration for the output widget."""

    format: str = config_field(
        default="%(levelname)s:%(message)s",
        tooltip="The logger format",
    )
    date_format: str = config_field(
        default="%Y-%m-%d %H:%M:%S",
        tooltip="The logger date format",
    )


@register_dock_widget_action(
    title="Output",
    area="right",
    menus=[MenuId.TOOLS_DOCK, MenuId.CORNER],
    keybindings=["Ctrl+Shift+U"],
    singleton=True,
    command_id="builtins:output",
    plugin_configs=OutputConfig(),
    icon="icon-park-outline:log",
)
def install_output_widget(ui: "MainWindow"):
    """Standard output widget."""
    from himena_builtins.qt.output._widget import get_widget

    return get_widget(ui.model_app.name)



from __future__ import annotations

from abc import abstractmethod
import inspect
from pathlib import Path
from typing import (
    Callable,
    Generic,
    TYPE_CHECKING,
    Hashable,
    Iterator,
    Literal,
    TypeVar,
)
from collections.abc import Sequence
import uuid
import math
import weakref

from psygnal import Signal
from himena import _providers
from himena._descriptors import SaveToPath
from himena.consts import ParametricWidgetProtocolNames as PWPN
from himena.layout import Layout, VBoxLayout, HBoxLayout, VStackLayout
from himena.plugins import _checker
from himena.types import (
    FutureInfo,
    Margins,
    NewWidgetBehavior,
    Size,
    WidgetDataModel,
    WindowState,
    WindowRect,
)
from himena.utils.collections import FrozenList
from himena.workflow import ProgrammaticMethod
from himena.widgets._wrapper import (
    _HasMainWindowRef,
    SubWindow,
    ParametricWindow,
    DockWidget,
)

if TYPE_CHECKING:
    from himena.widgets import BackendMainWindow
    from concurrent.futures import Future

    PathOrPaths = str | Path | list[str | Path]

_W = TypeVar("_W")  # backend widget type
_T = TypeVar("_T")  # type of the default value


class SemiMutableSequence(Sequence[_T]):
    @abstractmethod
    def __delitem__(self, i: int) -> None:
        return NotImplementedError

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)})"

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{type(self).__name__}(...)")
        else:
            p.text(f"{type(self).__name__}([\n")
            for item in self:
                p.text("  ")
                p.pretty(item)
                p.text(",\n")
            p.text("])")

    def clear(self):
        """Clear all the contents of the list."""
        for _ in range(len(self)):
            del self[-1]

    def remove(self, value: _T) -> None:
        """Remove the first occurrence of a value."""
        try:
            i = self.index(value)
        except ValueError:
            raise ValueError("Value not found in the list.")
        del self[i]

    def pop(self, index: int = -1):
        v = self[index]
        del self[index]
        return v

    def len(self) -> int:
        """Length of the list."""
        return len(self)

    def enumerate(self) -> Iterator[tuple[int, _T]]:
        """Method version of enumerate."""
        yield from enumerate(self)

    def iter(self) -> Iterator[_T]:
        """Method version of iter."""
        return iter(self)


class TabArea(SemiMutableSequence[SubWindow[_W]], _HasMainWindowRef[_W]):
    """An area containing multiple sub-windows."""

    def __init__(self, main_window: BackendMainWindow[_W], hash_value: Hashable):
        super().__init__(main_window)
        self._hash_value = hash_value
        # A tab area always has a layout for stacking minimized windows
        self._minimized_window_stack_layout = VStackLayout(main_window, inverted=True)
        self._layouts = [self._minimized_window_stack_layout]
        self._minimized_window_stack_layout._reanchor(Size(*main_window._area_size()))
        # the tab-specific result stack
        self._result_stack_ref: Callable[[], _W | None] = lambda: None

    @property
    def layouts(self) -> FrozenList[Layout]:
        """List of layouts in the tab area."""
        return FrozenList(self._layouts)

    def _tab_index(self) -> int:
        main = self._main_window()
        for i in range(main._num_tabs()):
            if main._tab_hash(i) == self._hash_value:
                return i
        raise ValueError("Tab is already removed from the main window.")

    def __getitem__(self, index_or_name: int | str) -> SubWindow[_W]:
        index = self._norm_index_or_name(index_or_name)
        widgets = self._main_window()._get_widget_list(self._tab_index())
        front = widgets[index][1]
        return front._himena_widget

    def __delitem__(self, index_or_name: int | str) -> None:
        index = self._norm_index_or_name(index_or_name)
        win, widget = self._pop_no_emit(index)
        _checker.call_widget_closed_callback(widget)
        win.closed.emit()

    def _pop_no_emit(self, index: int) -> tuple[SubWindow[_W], _W]:
        main = self._main_window()
        win = self[index]
        widget = win.widget  # get widget here to avoid garbage collection
        main._del_widget_at(self._tab_index(), index)
        main._remove_control_widget(widget)
        if isinstance(sb := win.save_behavior, SaveToPath):
            main._himena_main_window._history_closed.add((sb.path, sb.plugin))
        return win, widget

    def __len__(self) -> int:
        return len(self._main_window()._get_widget_list(self._tab_index()))

    def __iter__(self) -> Iterator[SubWindow[_W]]:
        return iter(
            w[1]._himena_widget
            for w in self._main_window()._get_widget_list(self._tab_index())
        )

    def append(self, sub_window: SubWindow[_W], title: str) -> None:
        """Append a sub-window to the tab area."""
        main = self._main_window()
        interf, front = sub_window._split_interface_and_frontend()
        front._himena_widget = sub_window
        out = main.add_widget(front, self._tab_index(), title)
        if hasattr(interf, "control_widget"):
            main._set_control_widget(front, interf.control_widget())

        main._connect_window_events(sub_window, out)
        sub_window.title = title
        sub_window.state_changed.connect(main._update_context)

        main._set_current_tab_index(self._tab_index())
        if main._himena_main_window._new_widget_behavior is NewWidgetBehavior.TAB:
            main._set_window_state(
                front,
                WindowState.FULL,
                main._himena_main_window._instructions.updated(animate=False),
            )

        main._move_focus_to(front)
        sub_window._alive = True
        return None

    def current(self, default: _T = None) -> SubWindow[_W] | _T:
        """Get the current sub-window or a default value."""
        idx = self.current_index
        if idx is None:
            return default
        try:
            return self[idx]
        except IndexError:
            return default

    @property
    def name(self) -> str:
        """Name of the tab area."""
        return self._main_window()._get_tab_name_list()[self._tab_index()]

    @property
    def current_index(self) -> int | None:
        """Get the index of the current sub-window."""
        return self._main_window()._current_sub_window_index(self._tab_index())

    @current_index.setter
    def current_index(self, index: int) -> None:
        self._main_window()._set_current_sub_window_index(self._tab_index(), index)

    @property
    def title(self) -> str:
        """Title of the tab area."""
        return self._main_window()._tab_title(self._tab_index())

    def add_widget(
        self,
        widget: _W,
        *,
        title: str | None = None,
        auto_size: bool = True,
    ) -> SubWindow[_W]:
        """Add a widget to the sub window.

        Parameters
        ----------
        widget : QtW.QWidget
            Widget to add.
        title : str, optional
            Title of the sub-window. If not given, its name will be automatically
            generated.

        Returns
        -------
        SubWindow
            A sub-window widget. The added widget is available by calling
            `widget` property.
        """
        main = self._main_window()
        sub_window = SubWindow(widget=widget, main_window=main)
        self._process_new_widget(sub_window, title, auto_size)
        main._move_focus_to(sub_window._split_interface_and_frontend()[1])
        _checker.call_theme_changed_callback(widget, main._himena_main_window.theme)
        return sub_window

    def add_function(
        self,
        func: Callable[..., _T],
        *,
        preview: bool = False,
        title: str | None = None,
        show_parameter_labels: bool = True,
        auto_close: bool = True,
        run_async: bool = False,
        result_as: Literal["window", "below", "right"] = "window",
    ) -> ParametricWindow[_W]:
        """Add a function as a parametric sub-window.

        The input function must return a `WidgetDataModel` instance, which can be
        interpreted by the application.

        Parameters
        ----------
        func : function (...) -> WidgetDataModel
            Function that generates a model from the input parameters.
        preview : bool, default False
            If true, the parametric widget will be implemented with a preview toggle
            button, and the preview window will be created when it is enabled.
        title : str, optional
            Title of the parametric window.
        show_parameter_labels : bool, default True
            If true, the parameter labels will be shown in the parametric window.
        auto_close : bool, default True
            If true, close the parametric window after the function call.

        Returns
        -------
        SubWindow
            The sub-window instance that represents the output model.
        """
        sig = inspect.signature(func)
        back_main = self._main_window()
        _is_prev_arg = ParametricWindow._IS_PREVIEWING
        if preview and _is_prev_arg in sig.parameters:
            parameters = [p for p in sig.parameters.values() if p.name != _is_prev_arg]
            sig = sig.replace(parameters=parameters)
        fn_widget = back_main._signature_to_widget(
            sig,
            show_parameter_labels=show_parameter_labels,
            preview=preview,
        )
        param_widget = self.add_parametric_widget(
            fn_widget, func, title=title, preview=preview, run_async=run_async,
            auto_close=auto_close, result_as=result_as,
        )  # fmt: skip
        return param_widget

    def add_parametric_widget(
        self,
        widget: _W,
        callback: Callable | None = None,
        *,
        title: str | None = None,
        preview: bool = False,
        auto_close: bool = True,
        auto_size: bool = True,
        run_async: bool = False,
        result_as: Literal["window", "below", "right"] = "window",
    ) -> ParametricWindow[_W]:
        """Add a custom parametric widget and its callback as a subwindow.

        This method creates a parametric window inside the workspace, so that the
        calculation can be done with the user-defined parameters.

        Parameters
        ----------
        widget : _W
            The parametric widget implemented with `get_params` and/or `get_output`.
        callback : callable, optional
            The callback function that will be called with the parameters set by the
            widget.
        title : str, optional
            Title of the window to manage parameters.
        preview : bool, default False
            If true, the parametric widget will be check for whether preview is enabled
            everytime the parameter changed, and if preview is enabled, a preview window
            is created to show the preview result.
        auto_close : bool, default True
            If true, close the parametric window after the function call.
        auto_size : bool, default True
            If true, the output window will be auto-sized to the size of the parametric
            window.

        Returns
        -------
        ParametricWindow[_W]
            A wrapper containing the backend widget.
        """
        if callback is None:
            if not hasattr(widget, PWPN.GET_OUTPUT):
                raise TypeError(
                    f"Parametric widget must have `{PWPN.GET_OUTPUT}` method if "
                    "callback is not given."
                )
            callback = getattr(widget, PWPN.GET_OUTPUT)
        main = self._main_window()
        widget0 = main._process_parametric_widget(widget)
        param_widget = ParametricWindow(widget0, callback, main_window=main)
        param_widget._auto_close = auto_close
        param_widget._result_as = result_as
        param_widget._run_asynchronously = run_async
        main._connect_parametric_widget_events(param_widget, widget0)
        self._process_new_widget(param_widget, title, auto_size)
        if preview:
            if not (
                hasattr(widget, PWPN.CONNECT_CHANGED_SIGNAL)
                and hasattr(widget, PWPN.IS_PREVIEW_ENABLED)
            ):
                raise TypeError(
                    f"If preview=True, the backend widget {widget!r} must implements "
                    f"methods {PWPN.CONNECT_CHANGED_SIGNAL!r} and "
                    f"{PWPN.IS_PREVIEW_ENABLED!r}"
                )
            param_widget.params_changed.connect(param_widget._widget_preview_callback)
        main._move_focus_to(widget0)
        return param_widget

    def add_layout(self, layout: Layout) -> Layout:
        layout._main_window_ref = weakref.ref(self._main_window())
        return self._add_layout_impl(layout)

    def add_vbox_layout(
        self,
        *,
        margins: Margins[int] | tuple[int, int, int, int] = (0, 0, 0, 0),
        spacing: int = 0,
    ) -> VBoxLayout:
        """Add a vertical box layout to the tab area.

        Parameters
        ----------
        margins : (int, int, int, int) or Margins, optional
            Left, top, right and bottom margins of the layout.
        spacing : int, optional
            Spacing between the widgets.
        """
        main = self._main_window()
        layout = VBoxLayout(main, margins=margins, spacing=spacing)
        return self._add_layout_impl(layout)

    def add_hbox_layout(
        self,
        *,
        margins: Margins[int] | tuple[int, int, int, int] = (0, 0, 0, 0),
        spacing: int = 0,
    ) -> HBoxLayout:
        """Add a horizontal box layout to the tab area.

        Parameters
        ----------
        margins : (int, int, int, int) or Margins, optional
            Left, top, right and bottom margins of the layout.
        spacing : int, optional
            Spacing between the widgets.
        """
        main = self._main_window()
        layout = HBoxLayout(main, margins=margins, spacing=spacing)
        return self._add_layout_impl(layout)

    def _add_layout_impl(self, layout: Layout) -> Layout:
        self._layouts.append(layout)
        layout._reanchor(self._main_window()._area_size())
        return layout

    def _process_new_widget(
        self,
        sub_window: SubWindow[_W],
        title: str | None = None,
        auto_size: bool = True,
    ) -> None:
        """Add, resize, and set the focus to the new widget."""
        main = self._main_window()
        interf, front = sub_window._split_interface_and_frontend()
        if title is None:
            title = getattr(interf, "default_title", _make_title)(len(self))
        out = main.add_widget(front, self._tab_index(), title)
        if hasattr(interf, "control_widget"):
            main._set_control_widget(front, interf.control_widget())

        main._connect_window_events(sub_window, out)
        sub_window.title = title
        sub_window.state_changed.connect(main._update_context)

        main._set_current_tab_index(self._tab_index())
        nwindows = len(self)
        if main._himena_main_window._new_widget_behavior is NewWidgetBehavior.TAB:
            main._set_window_state(
                front,
                WindowState.FULL,
                main._himena_main_window._instructions.updated(animate=False),
            )
        else:
            i_tab = main._current_tab_index()
            main._set_current_sub_window_index(i_tab, len(self) - 1)
            if auto_size:
                left = 4 + 24 * (nwindows % 5)
                top = 4 + 24 * (nwindows % 5)
                if size_hint := sub_window.size_hint():
                    width, height = size_hint
                else:
                    _, _, width, height = sub_window.rect
                sub_window.rect = WindowRect(left, top, width, height)
        _checker.call_widget_added_callback(sub_window.widget)
        sub_window._alive = True
        return None

    def add_data_model(self, model: WidgetDataModel) -> SubWindow[_W]:
        """Add a widget data model as a widget."""
        if not isinstance(model, WidgetDataModel):
            raise TypeError(
                f"input model must be an instance of WidgetDataModel, got {model!r}"
            )
        if len(model.workflow) == 0:
            # this case may happen if this method was programatically called
            wf = ProgrammaticMethod(output_model_type=model.type).construct_workflow()
            model = model.model_copy(update={"workflow": wf})
        ui = self._main_window()._himena_main_window
        widget = ui._pick_widget(model)
        ui.set_status_tip(f"Data model {model.title!r} added.", duration=1)
        sub_win = self.add_widget(widget, title=model.title)
        sub_win._update_from_returned_model(model)
        if rect_factory := model.window_rect_override:
            rect = WindowRect.from_tuple(*rect_factory(sub_win.size))
            sub_win.rect = rect
        if model.extension_default is not None:
            sub_win._extension_default_fallback = model.extension_default
        return sub_win

    def read_file(
        self,
        file_path: PathOrPaths,
        plugin: str | None = None,
    ) -> SubWindow[_W]:
        """Read local file(s) and open as a new sub-window in this tab.

        Parameters
        ----------
        file_path : str or Path or list of them
            Path(s) to the file to read. If a list is given, they will be read as a
            group, not as separate windows.
        plugin : str, optional
            If given, reader provider will be searched with the plugin name. This value
            is usually the full import path to the reader provider function, such as
            `"himena_builtins.io.default_reader_provider"`.

        Returns
        -------
        SubWindow
            The sub-window instance that is constructed based on the return value of
            the reader.
        """
        return self.read_files([file_path], plugin=plugin)[0]

    def read_files(
        self,
        file_paths: PathOrPaths,
        plugin: str | None = None,
    ) -> list[SubWindow[_W]]:
        """Read multiple files and open as new sub-windows in this tab."""
        models = self._paths_to_models(file_paths, plugin=plugin)
        out = [self.add_data_model(model) for model in models]
        ui = self._main_window()._himena_main_window
        if len(out) == 1:
            ui.set_status_tip(f"File opened: {out[0].title}", duration=5)
        elif len(out) > 1:
            _titles = ", ".join(w.title for w in out)
            ui.set_status_tip(f"File opened: {_titles}", duration=5)
        return out

    def _paths_to_models(self, file_paths: PathOrPaths, plugin: str | None = None):
        ins = _providers.ReaderStore.instance()
        file_paths = _norm_paths(file_paths)
        reader_path_sets = [
            (ins.pick(file_path, plugin=plugin), file_path) for file_path in file_paths
        ]
        models = [
            reader.read_and_update_source(file_path)
            for reader, file_path in reader_path_sets
        ]
        ui = self._main_window()._himena_main_window
        ui._recent_manager.append_recent_files(
            [(fp, reader.plugin_str) for reader, fp in reader_path_sets]
        )
        return models

    def read_files_async(
        self,
        file_paths: PathOrPaths,
        plugin: str | None = None,
    ) -> Future:
        """Read multiple files asynchronously and return a future."""
        ui = self._main_window()._himena_main_window
        file_paths = _norm_paths(file_paths)
        future = ui._executor.submit(self._paths_to_models, file_paths, plugin=plugin)
        if len(file_paths) == 1:
            ui.set_status_tip(f"Opening: {file_paths[0].as_posix()}", duration=5)
        else:
            ui.set_status_tip(f"Opening {len(file_paths)} files", duration=5)
        FutureInfo(list[WidgetDataModel]).set(future)  # set info for injection store
        return future

    def save_session(
        self,
        file_path: str | Path,
        save_copies: bool = False,
        allow_calculate: Sequence[str] = (),
    ) -> None:
        """Save the current session to a file."""
        from himena.session import dump_tab_to_zip

        dump_tab_to_zip(
            self, file_path, save_copies=save_copies, allow_calculate=allow_calculate
        )
        return None

    def tile_windows(
        self,
        nrows: int | None = None,
        ncols: int | None = None,
    ) -> None:
        main = self._main_window()
        inst = main._himena_main_window._instructions
        width, height = main._area_size()
        nrows, ncols = _norm_nrows_ncols(nrows, ncols, len(self))

        w = width / ncols
        h = height / nrows
        for i in range(nrows):
            for j in range(ncols):
                idx = i * ncols + j
                if idx >= len(self):
                    break
                x = j * width / ncols
                y = i * height / nrows
                sub = self[idx]
                rect = WindowRect.from_tuple(x, y, w, h)
                main._set_window_rect(sub.widget, rect, inst)
        return None

    def _norm_index_or_name(self, index_or_name: int | str) -> int:
        if isinstance(index_or_name, str):
            for i, w in enumerate(
                self._main_window()._get_widget_list(self._tab_index())
            ):
                if w[0] == index_or_name:
                    index = i
                    break
            else:
                raise ValueError(f"Name {index_or_name!r} not found.")
        else:
            if index_or_name < 0:
                index = len(self) + index_or_name
            else:
                index = index_or_name
        return index

    def _discard_result_stack_ref(self):
        """Discard the result stack reference."""
        self._result_stack_ref = lambda: None


def _norm_nrows_ncols(nrows: int | None, ncols: int | None, n: int) -> tuple[int, int]:
    if nrows is None:
        if ncols is None:
            nrows = int(math.sqrt(n))
            ncols = int(math.ceil(n / nrows))
        else:
            nrows = int(math.ceil(n / ncols))
    elif ncols is None:
        ncols = int(math.ceil(n / nrows))
    return nrows, ncols


class TabList(SemiMutableSequence[TabArea[_W]], _HasMainWindowRef[_W], Generic[_W]):
    """List of tab areas in the main window."""

    changed = Signal()

    def __init__(self, main_window: BackendMainWindow[_W]):
        super().__init__(main_window)
        self._tab_areas: dict[Hashable, TabArea] = {}

    def __getitem__(self, index_or_name: int | str) -> TabArea[_W]:
        index = self._norm_index_or_name(index_or_name)
        main = self._main_window()
        out = self._tab_areas.get(main._tab_hash(index))
        if out is None:
            raise ValueError(f"Tab index {index} not found in the list.")
        return out

    def __delitem__(self, index_or_name: int | str) -> None:
        index = self._norm_index_or_name(index_or_name)
        area = self[index]
        area.clear()
        main = self._main_window()
        hash = main._tab_hash(index)
        self._tab_areas.pop(hash)
        main._del_tab_at(index)
        self.changed.emit()
        return None

    def add(self, name: str) -> TabArea[_W]:
        """Add a new tab area with the given name."""
        main = self._main_window()
        n_tab = len(self)
        if name is None:
            name = f"Tab-{n_tab}"
        main.add_tab(name)
        self.changed.emit()
        main._set_current_tab_index(n_tab)
        area = TabArea(main, main._tab_hash(n_tab))
        self._tab_areas[area._hash_value] = area
        main._himena_main_window._tab_activated(n_tab)
        return area

    def get(self, index_or_name: int | str, /) -> TabArea[_W] | None:
        try:
            return self[index_or_name]
        except ValueError:
            return None

    def __len__(self) -> int:
        return len(self._main_window()._get_tab_name_list())

    def __iter__(self):
        main = self._main_window()
        for i in range(main._num_tabs()):
            hash = main._tab_hash(i)
            yield self._tab_areas[hash]

    def _norm_index_or_name(self, index_or_name: int | str) -> int:
        if isinstance(index_or_name, str):
            index = self.names.index(index_or_name)
        else:
            if index_or_name < 0:
                index = len(self) + index_or_name
            else:
                index = index_or_name
        return index

    @property
    def names(self) -> list[str]:
        """List of names of the tabs."""
        return self._main_window()._get_tab_name_list()

    def current(self, default: _T = None) -> TabArea[_W] | _T:
        """Get the current tab or a default value."""
        idx = self.current_index
        if idx is None:
            return default
        try:
            return self[idx]
        except (IndexError, ValueError):
            return default

    @property
    def current_index(self) -> int | None:
        """Get the index of the current tab (None of nothing exists)."""
        return self._main_window()._current_tab_index()

    @current_index.setter
    def current_index(self, index: int):
        self._main_window()._set_current_tab_index(index)


class DockWidgetList(
    SemiMutableSequence[DockWidget[_W]], _HasMainWindowRef[_W], Generic[_W]
):
    def __init__(self, main_window: BackendMainWindow[_W]):
        super().__init__(main_window)
        self._dock_widget_set = weakref.WeakValueDictionary[DockWidget[_W], _W]()

    def __getitem__(self, index: int) -> DockWidget[_W]:
        return list(self._dock_widget_set.keys())[index]

    def __delitem__(self, index: int) -> None:
        dock = self[index]
        self._main_window()._del_dock_widget(dock.widget)
        self._dock_widget_set.pop(dock)

    def __len__(self) -> int:
        return len(self._dock_widget_set)

    def __iter__(self) -> Iterator[DockWidget[_W]]:
        return iter(self._dock_widget_set.keys())

    def __repr__(self) -> str:
        return f"{type(self).__name__}({list(self)})"

    def _add_dock_widget(self, dock: DockWidget[_W]) -> None:
        self._dock_widget_set[dock] = dock.widget
        return None

    def widget_for_id(self, id: uuid.UUID) -> DockWidget[_W] | None:
        """Pick the dock widget by its ID."""
        for _dock in self:
            if id != _dock._identifier:
                continue
            return _dock
        return None


def _make_title(i: int) -> str:
    return f"Untitled-{i}"


def _norm_paths(file_paths: PathOrPaths) -> list[Path | list[Path]]:
    if isinstance(file_paths, (str, Path)):
        file_paths = [file_paths]
    out = []
    for file_path in file_paths:
        if isinstance(file_path, (str, Path)):
            file_path = Path(file_path)
        else:
            file_path = [Path(p) for p in file_path]
        out.append(file_path)
    return out



import numpy as np
from pytestqt.qtbot import QtBot
from himena import MainWindow, StandardType
from himena.testing import WidgetTester
from himena_builtins.qt.stack import QExcelEdit, QDataFrameStack, QArrayStack

def test_excel_widget(qtbot: QtBot, himena_ui: MainWindow):
    # himena_ui is needed because drop_model needs current instance
    excel_edit = QExcelEdit()
    with WidgetTester(excel_edit) as tester:
        qtbot.addWidget(excel_edit)
        excel_edit.show()
        tester.update_model(
            value={
                "sheet-0": {"a": [1, 2]},
                "sheet-1": {"a": [3, 4], "b": [5, 4]},
            },
        )
        old, new = tester.cycle_model()
        assert list(old.value.keys()) == list(new.value.keys())
        assert all(np.all(a == b) for a, b in zip(old.value.values(), new.value.values()))
        excel_edit.add_new_tab()
        assert excel_edit.count() == 3
        tester.drop_model(
            value={
                "sheet-10": {"a": [1, 2]},
                "sheet-11": [[1, 2], ["g", "g"]],
            },
            type=StandardType.EXCEL,
        )
        assert excel_edit.count() == 5
        tester.drop_model(
            value={"a": [1, 2]},
            type=StandardType.TABLE,
        )
        assert excel_edit.count() == 6

        control = tester.widget.control_widget()
        control._value_line_edit.setText("abc")
        control.update_for_editing()
        control._insert_row_above()
        control._insert_row_below()
        control._insert_column_right()
        control._insert_column_left()
        control._remove_selected_rows()
        control._remove_selected_columns()

    tabbar = excel_edit.tabBar()
    excel_edit.setCurrentIndex(0)
    tabbar._make_drag()

def test_dataframe_dict(qtbot: QtBot):
    widget = QDataFrameStack()
    with WidgetTester(widget) as tester:
        qtbot.addWidget(widget)
        widget.show()
        tester.update_model(
            value={
                "sheet-0": {"a": [1, 2]},
                "sheet-1": {"a": [3, 4], "b": [5, 4]},
            },
        )
        tester.cycle_model()
        tester.widget.control_widget()

def test_array_dict(qtbot: QtBot):
    widget = QArrayStack()
    with WidgetTester(widget) as tester:
        qtbot.addWidget(widget)
        widget.show()
        tester.update_model(
            value={
                "sheet-0": np.array([[1, 2], [3, 4]]),
                "sheet-1": np.array([[5, 6], [7, 8]]),
            },
        )
        tester.cycle_model()
        tester.widget.control_widget()

def test_command(himena_ui: MainWindow):
    himena_ui.add_object({"sheet-0": [[1, 2], [3, 4]]}, type=StandardType.EXCEL)
    himena_ui.exec_action("builtins:dict:duplicate-tab")
    assert himena_ui.current_model.type == StandardType.TABLE
    val = himena_ui.current_model.value
    assert val.tolist() == [["1", "2"], ["3", "4"]]



from __future__ import annotations

import logging
import traceback
from importlib import import_module
from typing import TYPE_CHECKING
from pathlib import Path
from timeit import default_timer as timer
from app_model.types import KeyBindingRule
from dataclasses import dataclass, field

if TYPE_CHECKING:
    from himena.profile import AppProfile
    from himena._app_model import HimenaApplication

_LOGGER = logging.getLogger(__name__)


def install_plugins(
    app: HimenaApplication, plugins: list[str]
) -> list[PluginInstallResult]:
    """Install plugins to the application."""
    from himena.plugins import AppActionRegistry
    from himena.profile import load_app_profile

    reg = AppActionRegistry.instance()
    results = []
    show_import_time = app.attributes.get("print_import_time", False)
    if show_import_time:
        print("==================")
        print("Plugin import time")
        print("==================")
    for name in plugins:
        if name in reg._installed_plugins:
            continue
        _time_0 = timer()
        _exc = None
        if isinstance(name, str):
            if name.endswith(".py"):
                if not Path(name).exists():
                    _LOGGER.error(
                        f"Plugin file {name} does not exists but is listed in the "
                        "application profile."
                    )
                    continue
                import runpy

                runpy.run_path(name)
            else:
                try:
                    import_module(name)
                except ModuleNotFoundError:
                    _LOGGER.error(
                        f"Plugin module {name} is not installed but is listed in the "
                        "application profile."
                    )
                    continue
                except Exception as e:
                    msg = "".join(
                        traceback.format_exception(type(e), e, e.__traceback__)
                    )
                    _LOGGER.error(
                        f"Error installing plugin {name}, traceback follows:\n{msg}"
                    )
                    _exc = e
        else:
            raise TypeError(f"Invalid plugin type: {type(name)}")
        _msec = (timer() - _time_0) * 1000
        if show_import_time and _exc is None:
            color = _color_for_time(_msec)
            print(f"{color}{name}\t{_msec:.3f} msec\033[0m")
        results.append(PluginInstallResult(name, _msec, _exc))
    reg.install_to(app)
    reg._installed_plugins.extend(plugins)
    prof = load_app_profile(app.name)

    for k, cfg in reg._plugin_default_configs.items():
        prof.plugin_configs.setdefault(k, cfg.as_dict())

    prof.save()
    return results


def override_keybindings(app: HimenaApplication, prof: AppProfile) -> None:
    """Override keybindings in the application."""
    for ko in prof.keybinding_overrides:
        if kb := app.keybindings.get_keybinding(ko.command_id):
            app.keybindings._keybindings.remove(kb)
        app.keybindings.register_keybinding_rule(
            ko.command_id,
            KeyBindingRule(primary=ko.key),
        )


@dataclass
class PluginInstallResult:
    plugin: str
    time: float = field(default=0.0)
    error: Exception | None = None


def _color_for_time(msec: float) -> str:
    """Return a color code for the given time in milliseconds."""
    if msec < 80:
        return "\033[92m"  # green
    elif msec < 700:
        return "\033[93m"  # yellow
    else:
        return "\033[91m"  # red



from __future__ import annotations
import math
from typing import TYPE_CHECKING
from psygnal import Signal
from qtpy import QtWidgets as QtW, QtCore, QtGui
from qtpy.QtCore import Qt

from ._roi_items import (
    QLineRoi,
    QPointRoi,
    QPointsRoi,
    QPolygonRoi,
    QRectangleRoi,
    QEllipseRoi,
    QCircleRoi,
    QRotatedRectangleRoi,
    QSegmentedLineRoi,
)

if TYPE_CHECKING:
    from ._graphics_view import QImageGraphicsView, QBaseGraphicsScene


class QHandleRect(QtW.QGraphicsRectItem):
    """The rect item for the ROI handles"""

    # NOTE: QGraphicsItem is not a QObject, so we can't use QtCore.Signal here
    moved_by_mouse = Signal(QtCore.QPointF, QtCore.QPointF)

    def __init__(self, x: int, y: int, width: int, height: int, parent=None):
        super().__init__(x, y, width, height, parent)
        self.setAcceptedMouseButtons(Qt.MouseButton.LeftButton)
        # this flag is needed to trigger mouseMoveEvent
        self.setFlag(QtW.QGraphicsItem.GraphicsItemFlag.ItemIsSelectable, True)
        self._pos_drag_start: QtCore.QPointF | None = None
        self._pos_drag_prev: QtCore.QPointF | None = None
        self._cursor_shape = Qt.CursorShape.PointingHandCursor
        self.setCursor(self._cursor_shape)
        self.setZValue(100000)

    def setColor(self, color: QtGui.QColor):
        self.setBrush(QtGui.QBrush(color))

    def center(self) -> QtCore.QPointF:
        """Return the center of the rect."""
        return self.rect().center()

    def setCenter(self, point: QtCore.QPointF):
        """Set the center of the rect."""
        x, y = point.x(), point.y()
        w, h = self.rect().width(), self.rect().height()
        self.setRect(x - w / 2, y - h / 2, w, h)

    def setSize(self, size: float):
        """Set the size of the rect."""
        x, y = self.center().x(), self.center().y()
        self.setRect(x - size / 2, y - size / 2, size, size)

    def translate(self, dx: float, dy: float):
        self.setRect(self.rect().translated(dx, dy))

    def mousePressEvent(self, event: QtW.QGraphicsSceneMouseEvent):
        grab = self.scene().grabSource()
        if grab is not None and grab is not self:
            return super().mousePressEvent(event)
        if event.button() == Qt.MouseButton.LeftButton:
            view = self.view()
            view.set_mode(self.view().Mode.SELECT)
            self.scene().setGrabSource(self)
        return super().mousePressEvent(event)

    def mouseMoveEvent(self, event: QtW.QGraphicsSceneMouseEvent | None) -> None:
        if self.scene().grabSource() is not self:
            return super().mouseMoveEvent(event)
        self.moved_by_mouse.emit(event.pos(), event.lastPos())
        return super().mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QtW.QGraphicsSceneMouseEvent | None) -> None:
        self.view().set_mode(self.view()._last_mode_before_key_hold)
        self.scene().setGrabSource(None)
        return super().mouseReleaseEvent(event)

    def view(self) -> QImageGraphicsView:
        return self.scene().views()[0]

    def scene(self) -> QBaseGraphicsScene:
        return super().scene()


class RoiSelectionHandles:
    draw_finished = Signal()

    def __init__(self, view: QImageGraphicsView):
        self._view = view
        self._handle_size = 2
        self._pen = QtGui.QPen(Qt.GlobalColor.black, 2)
        self._pen.setCosmetic(True)
        self._brush = QtGui.QBrush(Qt.GlobalColor.white)
        self._another_color = QtGui.QColor(Qt.GlobalColor.red)
        self._handles: list[QHandleRect] = []
        # This attribute is needed for drawing polygons and segmented lines.
        # When the mouse is hovering, the last vertice should not be considered as a
        # point yet, until the mouse is clicked.
        self._is_drawing_polygon = False
        self._is_last_vertex_added = False

    def make_handle_at(
        self,
        pos: QtCore.QPointF,
        color: QtGui.QColor | None = None,
    ) -> QHandleRect:
        """Construct a handle at the given position and add to the view."""
        s = self._handle_size
        handle = QHandleRect(pos.x() - s / 2, pos.y() - s / 2, s, s)
        handle.setPen(self._pen)
        handle.setBrush(self._brush)
        self._handles.append(handle)
        self.view().scene().addItem(handle)
        if color:
            handle.setColor(color)
        return handle

    def update_handle_size(self, scale: float):
        if scale == 0:
            return
        self._handle_size = 4 / scale
        for handle in self._handles:
            handle.setSize(self._handle_size)

    def clear_handles(self):
        """Remove all handles from the view."""
        view = self.view()
        for handle in self._handles:
            view.scene().removeItem(handle)
            handle.moved_by_mouse.disconnect()
        self._handles.clear()
        self.draw_finished.disconnect()

    def connect_line(self, line: QLineRoi):
        self.clear_handles()
        _line = line.line()
        h1 = self.make_handle_at(_line.p1())
        h2 = self.make_handle_at(_line.p2())
        hc = self.make_handle_at(_line.center(), self._another_color)
        self._handles = [h1, h2, hc]

        @h1.moved_by_mouse.connect
        def _1_moved(pos, last_pos):
            qline = line.line()
            qline.setP1(pos)
            line.setLine(qline)
            self._view.current_roi_updated.emit()

        @h2.moved_by_mouse.connect
        def _2_moved(pos, last_pos):
            qline = line.line()
            qline.setP2(pos)
            line.setLine(qline)
            self._view.current_roi_updated.emit()

        @hc.moved_by_mouse.connect
        def _c_moved(pos, last_pos):
            delta = pos - last_pos
            qline = line.line()
            qline.translate(delta.x(), delta.y())
            line.setLine(qline)
            self._view.current_roi_updated.emit()

        @line.changed.connect
        def _line_changed(qline: QtCore.QLineF):
            h1.setCenter(qline.p1())
            h2.setCenter(qline.p2())
            hc.setCenter(qline.center())
            self._view.current_roi_updated.emit()

    def connect_rect(self, rect: QRectangleRoi | QEllipseRoi):
        self.clear_handles()
        _rect = rect.rect()
        h_tl = self.make_handle_at(_rect.topLeft())
        h_br = self.make_handle_at(_rect.bottomRight())
        h_tr = self.make_handle_at(_rect.topRight())
        h_bl = self.make_handle_at(_rect.bottomLeft())
        h_t = self.make_handle_at(
            (_rect.topLeft() + _rect.topRight()) / 2, self._another_color
        )
        h_b = self.make_handle_at(
            (_rect.bottomLeft() + _rect.bottomRight()) / 2, self._another_color
        )
        h_l = self.make_handle_at(
            (_rect.topLeft() + _rect.bottomLeft()) / 2, self._another_color
        )
        h_r = self.make_handle_at(
            (_rect.topRight() + _rect.bottomRight()) / 2, self._another_color
        )
        self._handles = [h_tl, h_br, h_tr, h_bl, h_t, h_b, h_l, h_r]

        @h_tl.moved_by_mouse.connect
        def _tl_moved(pos, last_pos):
            other = rect.rect().bottomRight()
            ex, ey = self.view()._pos_to_tuple(pos)
            x0, x1 = sorted([other.x(), ex])
            y0, y1 = sorted([other.y(), ey])
            rect.setRect(x0, y0, x1 - x0, y1 - y0)
            self._view.current_roi_updated.emit()

        @h_br.moved_by_mouse.connect
        def _br_moved(pos, last_pos):
            other = rect.rect().topLeft()
            ex, ey = self.view()._pos_to_tuple(pos)
            x0, x1 = sorted([other.x(), ex])
            y0, y1 = sorted([other.y(), pos.y()])
            rect.setRect(x0, y0, x1 - x0, y1 - y0)
            self._view.current_roi_updated.emit()

        @h_tr.moved_by_mouse.connect
        def _tr_moved(pos, last_pos):
            other = rect.rect().bottomLeft()
            ex, ey = self.view()._pos_to_tuple(pos)
            x0, x1 = sorted([other.x(), ex])
            y0, y1 = sorted([other.y(), ey])
            rect.setRect(x0, y0, x1 - x0, y1 - y0)
            self._view.current_roi_updated.emit()

        @h_bl.moved_by_mouse.connect
        def _bl_moved(pos, last_pos):
            other = rect.rect().topRight()
            ex, ey = self.view()._pos_to_tuple(pos)
            x0, x1 = sorted([other.x(), ex])
            y0, y1 = sorted([other.y(), ey])
            rect.setRect(x0, y0, x1 - x0, y1 - y0)
            self._view.current_roi_updated.emit()

        @h_t.moved_by_mouse.connect
        def _t_moved(pos, last_pos):
            r0 = rect.rect()
            _, ey = self.view()._pos_to_tuple(pos)
            y0, y1 = sorted([r0.bottom(), ey])
            rect.setRect(r0.x(), y0, r0.width(), y1 - y0)
            self._view.current_roi_updated.emit()

        @h_b.moved_by_mouse.connect
        def _b_moved(pos, last_pos):
            r0 = rect.rect()
            _, ey = self.view()._pos_to_tuple(pos)
            y0, y1 = sorted([r0.top(), ey])
            rect.setRect(r0.x(), y0, r0.width(), y1 - y0)
            self._view.current_roi_updated.emit()

        @h_l.moved_by_mouse.connect
        def _l_moved(pos, last_pos):
            r0 = rect.rect()
            ex, _ = self.view()._pos_to_tuple(pos)
            x0, x1 = sorted([r0.right(), ex])
            rect.setRect(x0, r0.y(), x1 - x0, r0.height())
            self._view.current_roi_updated.emit()

        @h_r.moved_by_mouse.connect
        def _r_moved(pos, last_pos):
            r0 = rect.rect()
            ex, _ = self.view()._pos_to_tuple(pos)
            x0, x1 = sorted([r0.left(), ex])
            rect.setRect(x0, r0.y(), x1 - x0, r0.height())
            self._view.current_roi_updated.emit()

        @rect.changed.connect
        def _rect_changed(r: QtCore.QRectF):
            tl = r.topLeft()
            br = r.bottomRight()
            tr = r.topRight()
            bl = r.bottomLeft()
            h_tl.setCenter(tl)
            h_br.setCenter(br)
            h_tr.setCenter(tr)
            h_bl.setCenter(bl)
            h_t.setCenter((tl + tr) / 2)
            h_b.setCenter((bl + br) / 2)
            h_l.setCenter((tl + bl) / 2)
            h_r.setCenter((tr + br) / 2)
            self._view.current_roi_updated.emit()

    def connect_path(self, path: QPolygonRoi | QSegmentedLineRoi):
        self.clear_handles()
        _path = path.path()
        for i in range(_path.elementCount()):
            element = _path.elementAt(i)
            h = self.make_handle_at(QtCore.QPointF(element.x, element.y))
            h.moved_by_mouse.connect(lambda pos, _, i=i: path.update_point(i, pos))

        @path.changed.connect
        def _path_changed(p: QtGui.QPainterPath):
            offset = 0 if self._is_last_vertex_added else 1
            self.remove_handles(p.elementCount() - offset, len(self._handles))
            for i in range(len(self._handles), p.elementCount() - offset):
                element = p.elementAt(i)
                h = self.make_handle_at(QtCore.QPointF(element.x, element.y))
                h.moved_by_mouse.connect(lambda pos, _, i=i: path.update_point(i, pos))
            for i, h in enumerate(self._handles):
                element = p.elementAt(i)
                h.setCenter(QtCore.QPointF(element.x, element.y))
            self._view.current_roi_updated.emit()

        self.draw_finished.connect(lambda: self._finish_drawing_path(path))

    def connect_point(self, point: QPointRoi):
        self.clear_handles()
        h = self.make_handle_at(point.point())
        h.moved_by_mouse.connect(lambda pos, _: point.setPoint(pos))

        @point.changed.connect
        def _point_changed(ps: QtCore.QPointF):
            h.setCenter(ps)
            self._view.current_roi_updated.emit()

    def connect_points(self, points: QPointsRoi):
        self.clear_handles()
        for i in range(points.count()):
            h = self.make_handle_at(points.pointAt(i))
            h.moved_by_mouse.connect(lambda pos, _, i=i: points.update_point(i, pos))

        @points.changed.connect
        def _points_changed(ps: list[QtCore.QPointF]):
            self.remove_handles(len(ps), len(self._handles))
            for i in range(len(self._handles), len(ps)):
                h = self.make_handle_at(ps[i])
                h.moved_by_mouse.connect(
                    lambda pos, _, i=i: points.update_point(i, pos)
                )
            for i, h in enumerate(self._handles):
                h.setCenter(ps[i])
            self._view.current_roi_updated.emit()

    def connect_circle(self, circle: QCircleRoi):
        self.clear_handles()
        h_c = self.make_handle_at(circle.center())
        h_br = self.make_handle_at(circle.rect().bottomRight(), self._another_color)

        @h_c.moved_by_mouse.connect
        def _c_moved(pos: QtCore.QPointF, last_pos):
            circle.setCenterAndRadius((pos.x(), pos.y()), circle.radius())

        @h_br.moved_by_mouse.connect
        def _br_moved(pos: QtCore.QPointF, last_pos):
            radius = abs(pos.x() - circle.center().x())
            center = circle.center()
            circle.setCenterAndRadius((center.x(), center.y()), radius)

        @circle.changed.connect
        def _circle_changed(rect: QtCore.QRectF):
            h_c.setCenter(circle.center())
            h_br.setCenter(circle.rect().bottomRight())
            self._view.current_roi_updated.emit()

    def connect_rotated_rect(self, rect: QRotatedRectangleRoi):
        self.clear_handles()
        h_l = self.make_handle_at(rect.start())
        h_r = self.make_handle_at(rect.end())
        h_t = self.make_handle_at(rect.top(), self._another_color)
        h_b = self.make_handle_at(rect.bottom(), self._another_color)

        @h_t.moved_by_mouse.connect
        @h_b.moved_by_mouse.connect
        def _t_b_moved(pos, last_pos):
            ex = rect.vector_x()
            ex = ex / math.sqrt(ex.x() ** 2 + ex.y() ** 2)  # unit vector
            pos_rel = pos - rect.start()
            pos_proj = ex * QtCore.QPointF.dotProduct(pos_rel, ex)
            dist_vec = pos_rel - pos_proj
            dist = math.sqrt(dist_vec.x() ** 2 + dist_vec.y() ** 2)
            rect.set_width(dist * 2)
            self._view.current_roi_updated.emit()

        @h_l.moved_by_mouse.connect
        def _l_moved(pos, last_pos):
            rect.set_start(pos)
            self._view.current_roi_updated.emit()

        @h_r.moved_by_mouse.connect
        def _r_moved(pos, last_pos):
            rect.set_end(pos)
            self._view.current_roi_updated.emit()

        @rect.changed.connect
        def _rect_changed():
            h_l.setCenter(rect.start())
            h_r.setCenter(rect.end())
            h_t.setCenter(rect.top())
            h_b.setCenter(rect.bottom())
            self._view.current_roi_updated.emit()

    def connect_roi(self, roi_item):
        if isinstance(roi_item, QLineRoi):
            self.connect_line(roi_item)
        elif isinstance(roi_item, (QRectangleRoi, QEllipseRoi)):
            self.connect_rect(roi_item)
        elif isinstance(roi_item, (QPolygonRoi, QSegmentedLineRoi)):
            self.connect_path(roi_item)
        elif isinstance(roi_item, QPointRoi):
            self.connect_point(roi_item)
        elif isinstance(roi_item, QPointsRoi):
            self.connect_points(roi_item)
        elif isinstance(roi_item, QCircleRoi):
            self.connect_circle(roi_item)
        elif isinstance(roi_item, QRotatedRectangleRoi):
            self.connect_rotated_rect(roi_item)
        else:
            pass

    def _finish_drawing_path(self, path: QPolygonRoi | QSegmentedLineRoi):
        painter_path = path.path()
        if isinstance(path, QPolygonRoi):
            painter_path.closeSubpath()
        path.setPath(painter_path)

    def view(self) -> QImageGraphicsView:
        return self._view

    def start_drawing_polygon(self):
        self._is_drawing_polygon = True
        self._is_last_vertex_added = False

    def is_drawing_polygon(self) -> bool:
        return self._is_drawing_polygon

    def finish_drawing_polygon(self):
        was_drawing_polygon = self._is_drawing_polygon
        self._is_drawing_polygon = False
        if was_drawing_polygon:
            self.draw_finished.emit()

    def remove_handles(self, start: int, end: int):
        for i in range(start, end):
            self.view().scene().removeItem(self._handles[i])
        del self._handles[start:end]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar
from himena.utils.enum import StrEnum
from himena.types import Size, WindowRect, Margins

if TYPE_CHECKING:
    from himena.widgets import SubWindow

_W = TypeVar("_W", bound=Any)


class ResizeState(StrEnum):
    """The state of the resize operation of the window."""

    NONE = "none"
    TOP = "top"
    BOTTOM = "bottom"
    LEFT = "left"
    RIGHT = "right"
    TOP_LEFT = "top_left"
    TOP_RIGHT = "top_right"
    BOTTOM_LEFT = "bottom_left"
    BOTTOM_RIGHT = "bottom_right"

    @staticmethod
    def from_bools(
        is_left: bool, is_right: bool, is_top: bool, is_bottom: bool
    ) -> ResizeState:
        """Get the resize state from the edge booleans."""
        return RESIZE_STATE_MAP.get(
            (is_left, is_right, is_top, is_bottom), ResizeState.NONE
        )

    def resize_widget(
        self,
        widget_rect: WindowRect,
        mouse_pos: tuple[int, int],
        min_size: Size,
        max_size: Size,
    ) -> WindowRect | None:
        w_adj = _SizeAdjuster(min_size.width, max_size.width)
        h_adj = _SizeAdjuster(min_size.height, max_size.height)
        mx, my = mouse_pos
        if self is ResizeState.TOP_LEFT:
            out = WindowRect(
                widget_rect.left + mx,
                widget_rect.top + my,
                w_adj(widget_rect.width - mx),
                h_adj(widget_rect.height - my),
            )
        elif self is ResizeState.BOTTOM_LEFT:
            out = WindowRect(
                widget_rect.left + mx,
                widget_rect.top,
                w_adj(widget_rect.width - mx),
                h_adj(my),
            )
        elif self is ResizeState.TOP_RIGHT:
            out = WindowRect(
                widget_rect.left,
                widget_rect.top + my,
                w_adj(mx),
                h_adj(widget_rect.height - my),
            )
        elif self is ResizeState.BOTTOM_RIGHT:
            out = WindowRect(
                widget_rect.left,
                widget_rect.top,
                w_adj(mx),
                h_adj(my),
            )
        elif self is ResizeState.TOP:
            out = WindowRect(
                widget_rect.left,
                widget_rect.top + my,
                w_adj(widget_rect.width),
                h_adj(widget_rect.height - my),
            )
        elif self is ResizeState.BOTTOM:
            out = WindowRect(
                widget_rect.left,
                widget_rect.top,
                w_adj(widget_rect.width),
                h_adj(my),
            )
        elif self is ResizeState.LEFT:
            out = WindowRect(
                widget_rect.left + mx,
                widget_rect.top,
                w_adj(widget_rect.width - mx),
                h_adj(widget_rect.height),
            )
        elif self is ResizeState.RIGHT:
            out = WindowRect(
                widget_rect.left,
                widget_rect.top,
                w_adj(mx),
                h_adj(widget_rect.height),
            )
        else:
            out = None
        return out


class _SizeAdjuster:
    def __init__(self, min_x: int, max_x: int):
        self.min_x = min_x
        self.max_x = max_x

    def __call__(self, x: int) -> int:
        return min(max(x, self.min_x), self.max_x)


# is_left_edge, is_right_edge, is_top_edge, is_bottom_edge
RESIZE_STATE_MAP = {
    (True, False, True, False): ResizeState.TOP_LEFT,
    (False, True, True, False): ResizeState.TOP_RIGHT,
    (True, False, False, True): ResizeState.BOTTOM_LEFT,
    (False, True, False, True): ResizeState.BOTTOM_RIGHT,
    (True, False, False, False): ResizeState.LEFT,
    (False, True, False, False): ResizeState.RIGHT,
    (False, False, True, False): ResizeState.TOP,
    (False, False, False, True): ResizeState.BOTTOM,
    (False, False, False, False): ResizeState.NONE,
}


def prevent_window_overlap(
    win: SubWindow[_W],
    win_to_move: SubWindow[_W],
    area_size: tuple[int, int],
) -> WindowRect:
    """Find a comfortable position to move the window.

    Parameters
    ----------
    win : SubWindow
        The window that should not be overlapped by the other window.
    win_to_move : SubWindow
        The window that should be moved to prevent overlap with `win`.
    area_size : tuple[int, int]
        The size of the area in which the window should be moved.
    """
    offset = 8
    asize = Size(*area_size)
    margins = Margins.from_rects(win.rect, WindowRect(0, 0, *asize))
    rect_orig = win_to_move.rect
    size_orig = rect_orig.size()
    if size_orig.width < margins.right:
        out = rect_orig.move_top_left(
            win.rect.right + offset,
            min(win.rect.top, max(asize.height - size_orig.height - offset, 0)),
        )
    elif size_orig.width < margins.left:
        out = rect_orig.move_top_right(
            win.rect.left - offset,
            min(win.rect.top, max(asize.height - size_orig.height - offset, 0)),
        )
    elif size_orig.height < margins.bottom:
        out = rect_orig.move_top_left(
            min(win.rect.left, max(asize.width - size_orig.width - offset, 0)),
            max(win.rect.bottom + offset, 0),
        )
    elif size_orig.height < margins.top:
        out = rect_orig.move_bottom_left(
            min(win.rect.left, max(asize.width - size_orig.width - offset, 0)),
            max(win.rect.top - offset, rect_orig.height),
        )
    elif margins.bottom < margins.right:
        out = rect_orig.move_top_left(win.rect.right + offset, win.rect.top)
    else:
        out = rect_orig.move_bottom_left(win.rect.left, win.rect.bottom + offset)
    return out



from ._session import AppSession, TabSession
from ._api import (
    update_from_directory,
    update_from_zip,
    dump_directory,
    dump_zip,
    dump_tab_to_directory,
    dump_tab_to_zip,
)

__all__ = [
    "AppSession",
    "TabSession",
    "update_from_directory",
    "update_from_zip",
    "dump_directory",
    "dump_zip",
    "dump_tab_to_directory",
    "dump_tab_to_zip",
]



from typing import TYPE_CHECKING, Mapping
import weakref

from qtpy import QtWidgets as QtW, QtCore, QtGui

from himena.standards.model_meta import DictMeta
from himena.qt._qrename import QTabRenameLineEdit
from himena.qt import drag_command
from himena.types import DropResult, Parametric, WidgetDataModel
from himena.consts import StandardType
from himena.plugins import validate_protocol, register_hidden_function

_CMD_MERGE_TAB = "builtins:QDictOfWidgetEdit:merge-tab"
_CMD_SELECT_TAB = "builtins:QDictOfWidgetEdit:select-tab"


class QRightClickableTabBar(QtW.QTabBar):
    right_clicked = QtCore.Signal(int)

    def __init__(self, parent: "QDictOfWidgetEdit") -> None:
        super().__init__(parent)
        self._last_right_clicked: int | None = None
        self._is_dragging = False
        self._parent_ref = weakref.ref(parent)

    def mousePressEvent(self, a0: QtGui.QMouseEvent | None) -> None:
        if a0 is not None and a0.button() == QtCore.Qt.MouseButton.RightButton:
            self._last_right_clicked = self.tabAt(a0.pos())
        return super().mousePressEvent(a0)

    def mouseMoveEvent(self, a0):
        if self._is_dragging:
            return super().mouseMoveEvent(a0)
        if (
            a0.modifiers() & QtCore.Qt.KeyboardModifier.ControlModifier
            and a0.buttons() & QtCore.Qt.MouseButton.LeftButton
        ) or (a0.buttons() & QtCore.Qt.MouseButton.MiddleButton):
            self._is_dragging = True
            if drag := self._make_drag():
                drag.exec()
                return
        return super().mouseMoveEvent(a0)

    def mouseReleaseEvent(self, a0: QtGui.QMouseEvent | None) -> None:
        if a0 is not None and a0.button() == QtCore.Qt.MouseButton.RightButton:
            if self.tabAt(a0.pos()) == self._last_right_clicked:
                self.right_clicked.emit(self._last_right_clicked)
        self._last_right_clicked = None
        self._is_dragging = False
        return super().mouseReleaseEvent(a0)

    def _make_drag(self) -> QtGui.QDrag | None:
        if (qexcel := self._parent_ref()) and (index := qexcel.currentIndex()) >= 0:
            return drag_command(
                source=qexcel,
                type=qexcel._model_type_component,
                command_id=_CMD_SELECT_TAB,
                with_params={
                    "index": qexcel.currentIndex(),
                    "model_type": qexcel._model_type_component,
                },
                exec=False,
                desc=qexcel.tabText(index),
            )


class QDictOfWidgetEdit(QtW.QTabWidget):
    def __init__(self):
        super().__init__()
        self.setTabBar(QRightClickableTabBar(self))
        self._is_editable = True
        self._model_type_component = StandardType.ANY
        self._model_type = StandardType.DICT
        self._extension_default: str | None = None
        self.currentChanged.connect(self._on_tab_changed)
        self._line_edit = QTabRenameLineEdit(self, allow_duplicate=False)

        # corner widget for adding new tab
        tb = QtW.QToolButton()
        tb.setText("+")
        tb.setFont(QtGui.QFont("Arial", 12, weight=15))
        tb.setToolTip("New Tab")
        tb.clicked.connect(self.add_new_tab)
        self.setCornerWidget(tb, QtCore.Qt.Corner.TopRightCorner)
        self.tabBar().right_clicked.connect(self._tabbar_right_clicked)

    def _default_widget(self) -> QtW.QWidget:
        raise NotImplementedError

    def _on_tab_changed(self, index: int):
        self.control_widget().update_for_component(self.widget(index))
        return None

    def _tabbar_right_clicked(self, index: int):
        if index < 0:  # Clicked on the empty space
            return
        else:  # Clicked on an existing tab
            menu = self._menu_for_tabbar_right_clicked(index)
            menu.exec()

    def _menu_for_tabbar_right_clicked(self, index: int) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        rename_action = menu.addAction("Rename Tab")
        rename_action.triggered.connect(lambda: self._line_edit.start_edit(index))
        delete_action = menu.addAction("Delete Tab")
        delete_action.triggered.connect(lambda: self.removeTab(index))
        return menu

    def add_new_tab(self):
        table = self._default_widget()
        self.addTab(table, f"Sheet-{self.count() + 1}")
        self.setCurrentIndex(self.count() - 1)
        self.control_widget().update_for_component(table)
        return None

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        if not isinstance(value := model.value, Mapping):
            raise ValueError(f"Expected a dict, got {type(value)}")
        self.clear()
        for tab_name, each in value.items():
            table = self._default_widget()
            table.update_model(
                WidgetDataModel(value=each, type=self._model_type_component)
            )
            self.addTab(table, str(tab_name))
        if self.count() > 0:
            self.setCurrentIndex(0)
            self.control_widget().update_for_component(self.widget(0))
        self._model_type = model.type
        self._extension_default = model.extension_default
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        index = self.currentIndex()
        models: dict[str, WidgetDataModel] = {
            self.tabText(i): self.widget(i).to_model() for i in range(self.count())
        }
        return WidgetDataModel(
            value={tab_name: model.value for tab_name, model in models.items()},
            type=self.model_type(),
            extension_default=self._extension_default,
            metadata=DictMeta(
                current_tab=self.tabText(index),
                child_meta={
                    tab_name: model.metadata for tab_name, model in models.items()
                },
            ),
        )

    @validate_protocol
    def control_widget(self) -> "QTabControl":
        raise NotImplementedError

    @validate_protocol
    def model_type(self):
        return self._model_type

    @validate_protocol
    def is_modified(self) -> bool:
        return any(self.widget(i).is_modified() for i in range(self.count()))

    @validate_protocol
    def set_modified(self, value: bool) -> None:
        for i in range(self.count()):
            self.widget(i).set_modified(value)

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 300

    @validate_protocol
    def is_editable(self) -> bool:
        return self._is_editable

    @validate_protocol
    def set_editable(self, value: bool) -> None:
        self._is_editable = value
        for i in range(self.count()):
            self.widget(i).set_editable(value)

    @validate_protocol
    def allowed_drop_types(self) -> list[str]:
        return [self._model_type, self._model_type_component]

    @validate_protocol
    def dropped_callback(self, model: WidgetDataModel) -> DropResult:
        return DropResult(
            delete_input=True,
            command_id=_CMD_MERGE_TAB,
            with_params={"incoming": model},
        )

    if TYPE_CHECKING:

        def tabBar(self) -> QRightClickableTabBar: ...


class QTabControl(QtW.QWidget):
    def update_for_component(self, widget: QtW.QWidget | None):
        raise NotImplementedError


@register_hidden_function(command_id=_CMD_SELECT_TAB)
def select_tab(model: WidgetDataModel) -> Parametric:
    def run(index: int, model_type: str) -> WidgetDataModel:
        d = dict(model.value)
        key = list(d.keys())[index]
        value = d[key]
        return WidgetDataModel(
            value=value,
            type=model_type,
            title=key,
        )

    return run


@register_hidden_function(command_id=_CMD_MERGE_TAB)
def merge_tab(model: WidgetDataModel) -> Parametric:
    def run(incoming: WidgetDataModel) -> WidgetDataModel:
        out = dict(model.value)
        if incoming.is_subtype_of(StandardType.DICT):
            for key, value in dict(incoming.value).items():
                _update_dict_no_duplicate(out, key, value)
        else:
            _update_dict_no_duplicate(out, incoming.title, incoming.value)
        return model.with_value(out)

    return run


def _update_dict_no_duplicate(dict_: dict, key: str, value):
    if key not in dict_:
        dict_[key] = value
        return dict_

    if "-" in key and key.rsplit("-")[-1].isdigit():
        prefix, num_str = key.rsplit("-", 1)
        num = int(num_str) + 1
    else:
        prefix = key
        num = 0
    while f"{prefix}-{num}" in dict_:
        num += 1
    dict_[f"{prefix}-{num}"] = value
    return dict_



from __future__ import annotations


from magicgui.widgets.bases import ButtonWidget
from superqt import QToggleSwitch
from magicgui.backends._qtpy.widgets import QBaseButtonWidget


class ToggleSwitch(ButtonWidget):
    """A toggle switch widget behaves like a check box."""

    def __init__(self, **kwargs):
        super().__init__(
            widget_type=QBaseButtonWidget,
            backend_kwargs={"qwidg": QToggleSwitch},
            **kwargs,
        )
        self.native: QToggleSwitch



from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtCore
from himena.profile import profile_dir

if TYPE_CHECKING:
    from himena.qt._qmain_window import QMainWindow


class QStatusBar(QtW.QStatusBar):
    def __init__(self, parent: QMainWindow):
        super().__init__(parent)
        self._corner_widget = QtW.QWidget(self)
        layout = QtW.QHBoxLayout(self._corner_widget)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        layout.setContentsMargins(4, 0, 4, 0)
        self._profile_btn = QtW.QPushButton(f"{parent._app.name}")
        self._profile_btn.setToolTip("Application profile")
        self._profile_btn.setObjectName("profileButton")
        self._profile_btn.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)
        self._profile_btn.clicked.connect(self._open_profile_info)
        layout.addWidget(self._profile_btn)

        self.layout().setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.layout().addWidget(self._corner_widget)
        self._profile_info: QProfileInfo | None = None

    def _open_profile_info(self) -> None:
        """Open the profile info."""
        self._profile_info = info = QProfileInfo(self._profile_btn.text())
        info.setParent(self, QtCore.Qt.WindowType.Popup)
        info.show()
        info.move(
            self._profile_btn.mapToGlobal(QtCore.QPoint(0, 0))
            - QtCore.QPoint(0, info.height())
        )


class QProfileInfo(QtW.QWidget):
    def __init__(self, current_profile_name: str):
        super().__init__()
        label_texts = []
        for path in profile_dir().iterdir():
            if path.stem == current_profile_name:
                label_texts.append(f"&gt; <b>{path.stem}</b>")
            else:
                label_texts.append(f"&nbsp;&nbsp; {path.stem}")
        label_text = "<br>".join(label_texts)
        label = QtW.QLabel(label_text)
        label.setTextFormat(QtCore.Qt.TextFormat.RichText)
        label.setTextInteractionFlags(
            QtCore.Qt.TextInteractionFlag.TextBrowserInteraction
        )
        layout = QtW.QVBoxLayout(self)
        layout.addWidget(label)



from __future__ import annotations

import sys
import re
import subprocess
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Literal
from qtpy import QtWidgets as QtW, QtCore, QtGui
from superqt.utils import thread_worker
from superqt import QIconifyIcon, QToggleSwitch

from himena.workflow import RemoteReaderMethod
from himena import _drag
from himena.consts import MonospaceFontFamily
from himena.types import DragDataModel, WidgetDataModel
from himena.utils.misc import lru_cache
from himena.utils.cli import local_to_remote
from himena.widgets import set_status_tip, notify
from himena.plugins import validate_protocol
from himena_builtins.qt.widgets._shared import labeled

if TYPE_CHECKING:
    from himena.style import Theme
    from himena_builtins.qt.explorer import FileExplorerSSHConfig
    from himena.qt import MainWindowQt


class QSSHRemoteExplorerWidget(QtW.QWidget):
    """A widget for exploring remote files via SSH.

    This widget will execute `ls`, `ssh` and `scp` commands to list, read and send
    files when needed. This widget accepts copy-and-paste drag-and-drop from the local
    file system, including the normal explorer dock widget and the OS file explorer.

    If you are using Windows, checking the "Use WSL" switch will forward all the
    subprocess commands to WSL.
    """

    on_ls = QtCore.Signal(object)

    def __init__(self, ui: MainWindowQt) -> None:
        super().__init__()
        self.setAcceptDrops(True)
        font = QtGui.QFont(MonospaceFontFamily)
        self._ui = ui
        self._host_edit = QtW.QLineEdit()
        self._host_edit.setFont(font)
        self._host_edit.setMaximumWidth(100)
        self._user_name_edit = QtW.QLineEdit()
        self._user_name_edit.setFont(font)
        self._user_name_edit.setMaximumWidth(80)
        self._port_edit = QtW.QLineEdit()
        self._port_edit.setFont(font)
        self._port_edit.setValidator(QtGui.QIntValidator(0, 65535))
        self._port_edit.setMaximumWidth(40)
        self._is_wsl_switch = QToggleSwitch()
        self._is_wsl_switch.setText("Use WSL")
        self._is_wsl_switch.setFixedHeight(24)
        self._is_wsl_switch.setChecked(False)
        self._is_wsl_switch.setVisible(sys.platform == "win32")
        self._is_wsl_switch.setToolTip(
            "Use WSL (Windows Subsystem for Linux) to access remote files. If \n "
            "checked, all the subprocess commands such as `ls` will be prefixed \n"
            "with `wsl -e`."
        )
        self._protocol_choice = QtW.QComboBox()
        self._protocol_choice.addItems(["rsync", "scp"])
        self._protocol_choice.setCurrentIndex(0)
        self._protocol_choice.setToolTip("Choose the protocol to send files.")

        self._show_hidden_files_switch = QToggleSwitch()
        self._show_hidden_files_switch.setText("Hidden Files")
        self._show_hidden_files_switch.setToolTip("Also show hidden files")
        self._show_hidden_files_switch.setFixedHeight(24)
        self._show_hidden_files_switch.setChecked(False)

        self._pwd_widget = QtW.QLineEdit()
        self._pwd_widget.setFont(font)
        self._pwd_widget.editingFinished.connect(self._on_pwd_edited)

        self._last_dir_btn = QtW.QPushButton("←")
        self._last_dir_btn.setFixedWidth(20)
        self._last_dir_btn.setToolTip("Back to last directory")

        self._up_one_btn = QtW.QPushButton("↑")
        self._up_one_btn.setFixedWidth(20)
        self._up_one_btn.setToolTip("Up one directory")
        self._refresh_btn = QtW.QPushButton("Refresh")
        self._refresh_btn.setFixedWidth(60)
        self._refresh_btn.setToolTip("Refresh current directory")

        self._conn_btn = QtW.QPushButton("Connect")
        self._conn_btn.setFixedWidth(60)
        self._conn_btn.setToolTip("Connect to the remote host with the given user name")

        self._file_list_widget = QRemoteTreeWidget(self)
        self._file_list_widget.itemActivated.connect(self._read_item_to_gui)
        self._file_list_widget.setFont(font)
        self._file_list_widget.item_copied.connect(self._copy_item_paths)
        self._file_list_widget.item_pasted.connect(self._send_files)

        self._pwd = Path("~")
        self._last_dir = Path("~")

        self._filter_widget = QFilterLineEdit(self)
        self._filter_widget.textChanged.connect(self._apply_filter)
        self._filter_widget.setVisible(False)

        layout = QtW.QVBoxLayout(self)

        hlayout0 = QtW.QHBoxLayout()
        hlayout0.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(hlayout0)
        hlayout0.addWidget(labeled("Host:", self._host_edit, label_width=30), 3)
        hlayout0.addWidget(labeled("User:", self._user_name_edit, label_width=30), 2)
        hlayout0.addWidget(labeled("Port:", self._port_edit, label_width=30), 2)

        hlayout1 = QtW.QHBoxLayout()
        hlayout1.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(hlayout1)
        hlayout1.addWidget(self._is_wsl_switch)
        hlayout1.addWidget(self._protocol_choice)
        hlayout1.addWidget(self._conn_btn)

        layout.addWidget(QSeparator())
        layout.addWidget(labeled("Path:", self._pwd_widget))

        hlayout2 = QtW.QHBoxLayout()
        hlayout2.setContentsMargins(0, 0, 0, 0)
        hlayout2.addWidget(self._last_dir_btn, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        hlayout2.addWidget(self._up_one_btn, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        hlayout2.addWidget(QtW.QWidget(), 100)  # spacer
        hlayout2.addWidget(self._show_hidden_files_switch)
        hlayout2.addWidget(self._refresh_btn, 0, QtCore.Qt.AlignmentFlag.AlignRight)
        layout.addLayout(hlayout2)
        layout.addWidget(self._filter_widget)
        layout.addWidget(self._file_list_widget)

        self._conn_btn.clicked.connect(lambda: self._set_current_path(Path("~")))
        self._refresh_btn.clicked.connect(lambda: self._set_current_path(self._pwd))
        self._last_dir_btn.clicked.connect(
            lambda: self._set_current_path(self._last_dir)
        )
        self._up_one_btn.clicked.connect(
            lambda: self._set_current_path(self._pwd.parent)
        )
        self._show_hidden_files_switch.toggled.connect(
            lambda: self._set_current_path(self._pwd)
        )
        self._light_background = True

    @validate_protocol
    def theme_changed_callback(self, theme: Theme) -> None:
        self._light_background = theme.is_light_background()

    def _set_current_path(self, path: Path):
        self._pwd_widget.setText(path.as_posix())
        self._file_list_widget.clear()
        worker = self._run_ls_command(path)
        worker.returned.connect(self._on_ls_done)
        worker.started.connect(lambda: self._set_busy(True))
        worker.finished.connect(lambda: self._set_busy(False))
        worker.start()
        set_status_tip("Obtaining the file content ...", duration=3.0)

    def _on_ls_done(self, items: list[QtW.QTreeWidgetItem]):
        self._file_list_widget.addTopLevelItems(items)
        for i in range(1, self._file_list_widget.columnCount()):
            self._file_list_widget.resizeColumnToContents(i)
        set_status_tip(f"Currently under {self._pwd.name}", duration=1.0)

    def _set_busy(self, busy: bool):
        self._conn_btn.setEnabled(not busy)
        self._refresh_btn.setEnabled(not busy)
        self._last_dir_btn.setEnabled(not busy)
        self._up_one_btn.setEnabled(not busy)
        self._show_hidden_files_switch.setEnabled(not busy)
        self._file_list_widget.setEnabled(not busy)
        self._pwd_widget.setEnabled(not busy)

    def _host_name(self) -> str:
        username = self._user_name_edit.text()
        host = self._host_edit.text()
        return f"{username}@{host}"

    @thread_worker
    def _run_ls_command(self, path: Path) -> list[QtW.QTreeWidgetItem]:
        opt = "-lhAF" if self._show_hidden_files_switch.isChecked() else "-lhF"
        args = _make_ls_args(
            self._host_name(),
            path.as_posix(),
            options=opt,
            port=int(self._port_edit.text()),
        )
        if self._is_wsl_switch.isChecked():
            args = ["wsl", "-e"] + args
        result = subprocess.run(args, capture_output=True)
        if result.returncode != 0:
            raise ValueError(f"Failed to list directory: {result.stderr.decode()}")
        rows = result.stdout.decode().splitlines()
        # format of `ls -l` is:
        # <permission> <link> <owner> <group> <size> <month> <day> <time> <name>
        items: list[QtW.QTreeWidgetItem] = []
        for row in rows[1:]:  # the first line is total size
            *others, month, day, time, name = row.split(maxsplit=8)
            datetime = f"{month} {day} {time}"
            if name.endswith("*"):
                name = name[:-1]  # executable
            item = QtW.QTreeWidgetItem([name, datetime] + others[::-1])
            item.setToolTip(0, name)
            icon = _icon_for_file_type(_item_type(item), self._light_background)
            item.setIcon(0, icon)
            items.append(item)

        # sort directories first
        items = sorted(
            items,
            key=lambda x: (not x.text(0).endswith("/"), x.text(0)),
        )
        self._last_dir = self._pwd
        self._pwd = path
        return items

    def _read_item_to_gui(self, item: QtW.QTreeWidgetItem):
        item_type = _item_type(item)
        if item_type == "d":
            self._set_current_path(self._pwd / item.text(0))
        elif item_type == "l":
            _, real_path = item.text(0).split(" -> ")
            args_check_type = _make_get_type_args(
                self._host_name(), real_path, port=int(self._port_edit.text())
            )
            if self._is_wsl_switch.isChecked():
                args_check_type = ["wsl", "-e"] + args_check_type
            result = subprocess.run(args_check_type, capture_output=True)
            if result.returncode != 0:
                raise ValueError(f"Failed to get type: {result.stderr.decode()}")
            link_type = result.stdout.decode().strip()
            if link_type == "directory":
                self._set_current_path(self._pwd / real_path)
            else:
                self._read_and_add_model(self._pwd / real_path)
        else:
            self._read_and_add_model(self._pwd / item.text(0))

    def _copy_item_paths(self, items: list[QtW.QTreeWidgetItem]):
        mime = self._make_mimedata_for_items(items)
        clipboard = QtGui.QGuiApplication.clipboard()
        clipboard.setMimeData(mime)

    def _send_files(self, paths: list[Path]):
        for path in paths:
            self._ui.submit_async_task(self._send_file, path, path.is_dir())

    def _make_reader_method(self, path: Path, is_dir: bool) -> RemoteReaderMethod:
        return RemoteReaderMethod(
            host=self._host_edit.text(),
            username=self._user_name_edit.text(),
            path=path,
            port=int(self._port_edit.text()),
            wsl=self._is_wsl_switch.isChecked(),
            protocol=self._protocol_choice.currentText(),
            force_directory=is_dir,
        )

    def readers_from_mime(self, mime: QtCore.QMimeData) -> list[RemoteReaderMethod]:
        """Construct readers from the mime data."""
        is_wsl = self._is_wsl_switch.isChecked()
        prot = self._protocol_choice.currentText()
        out: list[RemoteReaderMethod] = []
        for line in mime.html().split("<br>"):
            if not line:
                continue
            if m := re.compile(r"<span ftype=\"(d|f)\">(.+)</span>").match(line):
                is_dir = m.group(1) == "d"
                line = m.group(2)
            else:
                continue
            meth = RemoteReaderMethod.from_str(
                line,
                wsl=is_wsl,
                protocol=prot,
                force_directory=is_dir,
            )
            out.append(meth)
        return out

    def _make_mimedata_for_items(
        self,
        items: list[QtW.QTreeWidgetItem],
    ) -> QtCore.QMimeData:
        mime = QtCore.QMimeData()
        mime.setText(
            "\n".join(
                meth.to_str() for meth in self._make_reader_methods_for_items(items)
            )
        )
        mime.setHtml(
            "<br>".join(
                f'<span ftype="{"d" if meth.force_directory else "f"}">{meth.to_str()}</span>'
                for meth in self._make_reader_methods_for_items(items)
            )
        )
        mime.setParent(self)  # this is needed to trace where the MIME data comes from
        return mime

    def _make_reader_methods_for_items(
        self, items: list[QtW.QTreeWidgetItem]
    ) -> list[RemoteReaderMethod]:
        methods: list[RemoteReaderMethod] = []
        for item in items:
            item_type = _item_type(item)
            if item_type == "l":
                _, real_path = item.text(0).split(" -> ")
                remote_path = self._pwd / real_path
                is_dir = False
            else:
                remote_path = self._pwd / item.text(0)
                is_dir = item_type == "d"
            meth = self._make_reader_method(remote_path, is_dir)
            methods.append(meth)
        return methods

    @thread_worker
    def _read_remote_path_worker(
        self, path: Path, is_dir: bool = False
    ) -> WidgetDataModel:
        return self._make_reader_method(path, is_dir).run()

    def _read_and_add_model(self, path: Path, is_dir: bool = False):
        """Read the remote file in another thread and add the model in the main."""
        worker = self._read_remote_path_worker(path, is_dir)
        worker.returned.connect(self._ui.add_data_model)
        worker.started.connect(lambda: self._set_busy(True))
        worker.finished.connect(lambda: self._set_busy(False))
        worker.start()
        set_status_tip(f"Reading file: {path}", duration=2.0)

    def _on_pwd_edited(self):
        pwd_text = self._pwd_widget.text()
        if "*" in pwd_text or "?" in pwd_text:
            self._pwd_widget.setSelection(0, len(pwd_text))
            raise ValueError("Wildcards are not supported.")
        if self._pwd != Path(pwd_text):
            self._set_current_path(Path(pwd_text))

    def dragEnterEvent(self, a0):
        if _drag.get_dragging_model() is not None or a0.mimeData().urls():
            a0.accept()
        else:
            a0.ignore()

    def dragMoveEvent(self, a0):
        a0.acceptProposedAction()
        return super().dragMoveEvent(a0)

    def dropEvent(self, a0):
        if model := _drag.drop():
            self._ui.submit_async_task(self._send_model, model)
            set_status_tip("Start sending file ...")
        elif urls := a0.mimeData().urls():
            for url in urls:
                path = Path(url.toLocalFile())
                self._ui.submit_async_task(self._send_file, path, path.is_dir())
                set_status_tip(f"Sent to {self._host_name()}:{path.name}", duration=2.8)

    def update_configs(
        self,
        cfg: FileExplorerSSHConfig,
    ) -> None:
        self._host_edit.setText(cfg.default_host)
        self._user_name_edit.setText(cfg.default_user)
        self._port_edit.setText(str(cfg.default_port))
        self._is_wsl_switch.setChecked(cfg.default_use_wsl)
        self._protocol_choice.setCurrentText(cfg.default_protocol)
        if cfg.default_host and cfg.default_user and self._pwd == Path("~"):
            self._set_current_path(Path("~"))

    def _send_model(self, model: DragDataModel):
        data_model = model.data_model()
        with tempfile.TemporaryDirectory() as tmpdir:
            tmpdir = Path(tmpdir)
            src_pathobj = data_model.write_to_directory(tmpdir)
            self._send_file(src_pathobj)

    def _send_file(self, src: Path, is_dir: bool = False):
        """Send local file to the remote host."""
        dst_remote = self._pwd / src.name
        args = local_to_remote(
            self._protocol_choice.currentText(),
            src,
            f"{self._host_name()}:{dst_remote.as_posix()}",
            is_wsl=self._is_wsl_switch.isChecked(),
            is_dir=is_dir,
            port=int(self._port_edit.text()),
        )
        subprocess.run(args)
        notify(f"Sent {src.as_posix()} to {dst_remote.as_posix()}", duration=2.8)

    def _apply_filter(self, text: str):
        for i in range(self._file_list_widget.topLevelItemCount()):
            item = self._file_list_widget.topLevelItem(i)
            ok = all(part in item.text(0).lower() for part in text.lower().split(" "))
            item.setHidden(not ok)

    def keyPressEvent(self, a0):
        if (
            a0.key() == QtCore.Qt.Key.Key_F
            and a0.modifiers() == QtCore.Qt.KeyboardModifier.ControlModifier
        ):
            self._filter_widget.setVisible(not self._filter_widget.isVisible())
            self._filter_widget.setVisible(True)
            self._filter_widget.setFocus()
            return
        return super().keyPressEvent(a0)


class QFilterLineEdit(QtW.QLineEdit):
    def __init__(self, parent: QSSHRemoteExplorerWidget):
        super().__init__(parent)
        self.setPlaceholderText("Filter files...")

    def keyPressEvent(self, a0):
        if a0.key() == QtCore.Qt.Key.Key_Escape:
            self.clear()
            self.setVisible(False)
            return
        return super().keyPressEvent(a0)


class QRemoteTreeWidget(QtW.QTreeWidget):
    item_copied = QtCore.Signal(list)
    item_pasted = QtCore.Signal(list)  # list of local Path objects

    def __init__(self, parent: QSSHRemoteExplorerWidget):
        super().__init__(parent)
        self.setIndentation(0)
        self.setColumnWidth(0, 180)
        self.setHeaderLabels(
            ["Name", "Datetime", "Size", "Group", "Owner", "Link", "Permission"]
        )
        self.setSelectionMode(QtW.QAbstractItemView.SelectionMode.ExtendedSelection)
        self.header().setDefaultAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
        self.header().setFixedHeight(20)
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._show_context_menu)

    def _make_context_menu(self):
        menu = QtW.QMenu(self)
        open_action = menu.addAction("Open")
        open_action.triggered.connect(
            lambda: self.itemActivated.emit(self.currentItem(), 0)
        )
        copy_action = menu.addAction("Copy")
        copy_action.triggered.connect(
            lambda: self.item_copied.emit(self.selectedItems())
        )
        paste_action = menu.addAction("Paste")
        paste_action.triggered.connect(self._paste_from_clipboard)
        menu.addSeparator()
        download_action = menu.addAction("Download")
        download_action.triggered.connect(
            lambda: self._save_items(self.selectedItems())
        )
        return menu

    def _show_context_menu(self, pos: QtCore.QPoint):
        self._make_context_menu().exec(self.viewport().mapToGlobal(pos))

    def keyPressEvent(self, event):
        _mod = event.modifiers()
        _key = event.key()
        if _mod == QtCore.Qt.KeyboardModifier.ControlModifier:
            if _key == QtCore.Qt.Key.Key_C:
                items = self.selectedItems()
                self.item_copied.emit(items)
                return None
            elif _key == QtCore.Qt.Key.Key_V:
                return self._paste_from_clipboard()
        return super().keyPressEvent(event)

    def _paste_from_clipboard(self):
        clipboard = QtGui.QGuiApplication.clipboard()
        mime = clipboard.mimeData()
        if mime.hasUrls():
            urls = mime.urls()
            paths = [Path(url.toLocalFile()) for url in urls]
            self.item_pasted.emit(paths)
        else:
            notify("No valid file paths in the clipboard.")

    # drag-and-drop
    def mouseMoveEvent(self, e: QtGui.QMouseEvent):
        if e.buttons() & QtCore.Qt.MouseButton.LeftButton:
            self._start_drag(e.pos())
            return None
        return super().mouseMoveEvent(e)

    def _start_drag(self, pos: QtCore.QPoint):
        items = self.selectedItems()
        mime = self.parent()._make_mimedata_for_items(items)
        drag = QtGui.QDrag(self)
        drag.setMimeData(mime)
        drag.exec(QtCore.Qt.DropAction.CopyAction)

    def _save_items(self, items: list[QtW.QTreeWidgetItem]):
        """Save the selected items to local files."""
        download_dir = Path.home() / "Downloads"
        src_paths: list[Path] = []
        for item in items:
            item_type = _item_type(item)
            if item_type == "l":
                _, real_path = item.text(0).split(" -> ")
                remote_path = self.parent()._pwd / real_path
            else:
                remote_path = self.parent()._pwd / item.text(0)
            src_paths.append(remote_path)

        readers = self.parent()._make_reader_methods_for_items(items)
        worker = make_paste_remote_files_worker(readers, download_dir)
        qui = self.parent()._ui._backend_main_window
        qui._job_stack.add_worker(worker, "Downloading files", total=len(src_paths))
        worker.start()

    if TYPE_CHECKING:

        def parent(self) -> QSSHRemoteExplorerWidget: ...


def _make_ls_args(
    host: str, path: str, port: int = 22, options: str = "-AF"
) -> list[str]:
    return ["ssh", "-p", str(port), host, "ls", path + "/", options]


def _make_get_type_args(host: str, path: str, port: int = 22) -> list[str]:
    return ["ssh", "-p", str(port), host, "stat", path, "--format='%F'"]


def _item_type(item: QtW.QTreeWidgetItem) -> Literal["d", "l", "f"]:
    """First character of the permission string."""
    return item.text(6)[0]


@lru_cache(maxsize=10)
def _icon_for_file_type(file_type: str, light_background: bool) -> QIconifyIcon:
    color = "#222222" if light_background else "#eeeeee"
    if file_type == "d":
        icon = QIconifyIcon("material-symbols:folder-rounded", color=color)
    elif file_type == "l":
        icon = QIconifyIcon("octicon:file-directory-symlink-16", color=color)
    else:
        icon = QIconifyIcon("mdi:file-outline", color=color)
    return icon


class QSeparator(QtW.QFrame):
    def __init__(self):
        super().__init__()
        self.setFrameShape(QtW.QFrame.Shape.HLine)
        self.setFrameShadow(QtW.QFrame.Shadow.Sunken)
        self.setFixedHeight(2)
        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )


@thread_worker
def make_paste_remote_files_worker(
    readers: list[RemoteReaderMethod],
    dirpath: Path,
):
    for reader in readers:
        dst = dirpath / reader.path.name
        reader.run_command(dst)
        if dst.exists():
            dst.touch()
        yield



from himena.plugins import register_dock_widget_action, add_default_status_tip
from himena.consts import MenuId

add_default_status_tip(
    short="command history",
    long="Ctrl+Shift+H to check, copy and re-run executed commands",
)


@register_dock_widget_action(
    title="Command History",
    menus=[MenuId.TOOLS_DOCK, MenuId.CORNER],
    area="right",
    keybindings=["Ctrl+Shift+H"],
    singleton=True,
    command_id="builtins:command-history",
    icon="mdi:view-list",
)
def install_command_history(ui):
    """A command history widget for viewing and executing commands."""
    from himena_builtins.qt.history._widget import QCommandHistory

    return QCommandHistory(ui)



from __future__ import annotations

from typing import Any
from himena.consts import StandardType


class ObjectTypeMap:
    def __init__(self):
        self._functions = []

    def pick_type(self, value: Any) -> str:
        for func in self._functions:
            out = func(value)
            if isinstance(out, str):
                return out
        raise ValueError(f"Could not determine the type of {value}.")

    def register(self, func):
        self._functions.append(func)
        return func


def register_defaults(map: ObjectTypeMap):
    import numpy as np
    from himena.workflow import Workflow

    @map.register
    def str_as_text(value: Any) -> str | None:
        if isinstance(value, str):
            return StandardType.TEXT
        return None

    @map.register
    def as_array(value) -> str | None:
        if isinstance(value, np.ndarray):
            if value.ndim == 2 and isinstance(value.dtype, np.dtypes.StringDType):
                return StandardType.TABLE
            return StandardType.ARRAY
        return None

    @map.register
    def as_dataframe(value) -> str | None:
        if hasattr(value, "__dataframe__"):
            return StandardType.DATAFRAME
        return None

    @map.register
    def as_workflow(value) -> str | None:
        if isinstance(value, Workflow):
            return StandardType.WORKFLOW
        return None



from __future__ import annotations

from typing import Callable, Any


def _format_float(value, ndigits: int = 4) -> str:
    """convert string to int or float if possible"""
    if value is None:
        return "null"
    if 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
        text = f"{value:.{ndigits}f}"
    else:
        text = f"{value:.{ndigits-1}e}"

    return text


def _format_int(value, ndigits: int = 4) -> str:
    if value is None:
        return "null"
    if 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
        text = str(value)
    else:
        text = f"{value:.{ndigits-1}e}"

    return text


def _format_complex(value: complex, ndigits: int = 3) -> str:
    if value != value:  # nan
        text = "nan"
    elif 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
        text = f"{value.real:.{ndigits}f}{value.imag:+.{ndigits}f}j"
    else:
        text = f"{value.real:.{ndigits-1}e}{value.imag:+.{ndigits-1}e}j"

    return text


def _format_datetime(value):
    return str(value)


_DEFAULT_FORMATTERS: dict[int, Callable[[Any], str]] = {
    "i": _format_int,
    "u": _format_int,
    "f": _format_float,
    "c": _format_complex,
    "t": _format_datetime,
}


def format_table_value(value: Any, fmt: str) -> str:
    return _DEFAULT_FORMATTERS.get(fmt, str)(value)
